from .matching_resample import (
    MatchingResampled,
    MatchingResampleD,
    MatchingResampleDict
)
from .pet_to_suv import (
    PETtoSUVd,
    PETtoSUVD,
    PETtoSUVDict
)
from .resample import (
    Resampled,
    ResampleD,
    ResampleDict
)



"""
    @file:              tools.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 10/2022

    @Description:       This file contains some tools to help applying transforms on images.
"""

from enum import IntEnum
from typing import Union

from monai.data import MetaTensor
from monai.utils import convert_to_numpy as monai_convert_to_numpy
import numpy as np

from delia.utils.data_model import PatientDataModel


class Mode(IntEnum):
    NONE = -1
    IMAGE = 0
    SEGMENTATION = 1


def set_transforms_keys(patient_dataset: PatientDataModel) -> None:
    """
    Sets transforms keys.

    Parameters
    ----------
    patient_dataset : PatientDataModel
        Patient dataset.
    """
    keys = []
    for image_and_segmentation_data in patient_dataset.data:
        series_key = image_and_segmentation_data.image.series_key
        modality = image_and_segmentation_data.image.dicom_header.Modality

        if series_key and series_key in keys:
            raise AssertionError(f"Series key {series_key} already in dict. To use 'transforms', there must be only "
                                 f"one image associated to a given series key.")
        elif series_key:
            image_and_segmentation_data.image.transforms_key = series_key
            keys.append(series_key)
        elif modality in keys:
            raise AssertionError(f"Modality {modality} already in dict. To use 'transforms' without using "
                                 f"'series_descriptions', there must be only one image associated to a given modality. "
                                 f"Otherwise, use 'series_descriptions' to specify the images' names as keys in the"
                                 f"'series_descriptions' dictionary.")
        else:
            image_and_segmentation_data.image.transforms_key = modality
            keys.append(modality)


def convert_to_numpy(array: Union[MetaTensor, np.ndarray]) -> np.ndarray:
    """
    Converts given image tensor or array to numpy array.

    Parameters
    ----------
    array : Union[MetaTensor, np.ndarray]
        Image tensor or array.

    Returns
    -------
    array : np.ndarray
        Image numpy array.
    """
    if isinstance(array, MetaTensor):
        return monai_convert_to_numpy(array[0, :])
    else:
        return array



"""
    @file:              copy_segmentations.py
    @Author:            Maxence Larose

    @Creation Date:     11/2022
    @Last modification: 11/2022

    @Description:       This file contains the CopySegmentationsd transform.
"""

from typing import Dict, Hashable

import SimpleITK as sitk

from delia.transforms.data.transform import DataTransform
from delia.utils.data_model import ImageAndSegmentationDataModel, SegmentationDataModel


class CopySegmentationsd(DataTransform):
    """
    Copies all segmentations associated with an image and assigns the copies to another image. We also make sure to
    resample the label maps to the same size as the "new" image. For example, this makes it possible to associate
    the segmentations made on a CT image to a PET image that has been acquired during the same scan.
    """

    def __init__(
            self,
            segmented_image_key: str,
            unsegmented_image_key: str,
    ):
        """
        Initialize transform.

        Parameters
        ----------
        segmented_image_key : str
            Key of the segmented image from which to copy the segmentations. Image keys are assumed to be arbitrary
            series keys defined in 'series_descriptions'. Note that if 'series_descriptions' is None, the image keys
            are assumed to be modality names.
        unsegmented_image_key : str
            Key of the unsegmented image on which to copy the segmentations. Image keys are assumed to be arbitrary
            series keys defined in 'series_descriptions'. Note that if 'series_descriptions' is None, the image keys
            are assumed to be modality names.
        """
        super().__init__(keys=[segmented_image_key, unsegmented_image_key])

        self._segmented_image_key = segmented_image_key
        self._unsegmented_image_key = unsegmented_image_key

    def __call__(self, data: Dict[str, ImageAndSegmentationDataModel]) -> Dict[Hashable, ImageAndSegmentationDataModel]:
        """
        Copies all segmentations associated with an image and assigns the copies to another image.

        Parameters
        ----------
        data : Dict[str, ImageAndSegmentationDataModel]
            A Python dictionary that contains ImageAndSegmentationDataModel.

        Returns
        -------
        transformed_data : Dict[Hashable, ImageAndSegmentationDataModel]
            A Python dictionary that contains transformed ImageAndSegmentationDataModel.
        """
        d = dict(data)

        image, segmentations = None, None
        for key in self.key_iterator(d):
            if key == self._segmented_image_key:
                segmentations = d[key].segmentations
            if key == self._unsegmented_image_key:
                image = d[key].image

                if d[key].segmentations:
                    raise AssertionError(
                        f"'CopySegmentationsd' found segmentations associated to 'unsegmented_image' = "
                        f"{self._segmented_image_key}. Unsegmented images are supposed to be unsegmented."
                    )

        if not image:
            raise AssertionError(
                f"'CopySegmentationsd' found no image associated to 'unsegmented_image' = "
                f"{self._unsegmented_image_key}.")
        if not segmentations:
            raise AssertionError(
                f"'CopySegmentationsd' found no segmentations associated to 'segmented_image' = "
                f"{self._segmented_image_key}."
            )

        new_segmentations = []
        for segmentation in segmentations:
            new_label_maps = {}
            for organ, mask in segmentation.simple_itk_label_maps.items():
                resampled_mask = sitk.Resample(
                    mask,
                    image.simple_itk_image,
                    sitk.Transform(),
                    sitk.sitkNearestNeighbor,
                    0,
                    mask.GetPixelID()
                )

                new_label_maps[organ] = resampled_mask

            new_segmentations.append(
                SegmentationDataModel(
                    dicom_header=segmentation.dicom_header,
                    modality=segmentation.modality,
                    simple_itk_label_maps=new_label_maps
                )
            )

        d[self._unsegmented_image_key].segmentations = new_segmentations

        return d


CopySegmentationsD = CopySegmentationsDict = CopySegmentationsd



"""
    @file:              patient_data_factories.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 03/2022

    @Description:       This file contains all factories that inherit from the BasePatientDataFactory class.
"""

from typing import Dict, List, Optional

from delia.readers.patient_data.factories.base_patient_data_factory import BasePatientDataFactory
from delia.readers.image.dicom_reader import DicomReader
from delia.readers.segmentation.segmentation_reader import SegmentationReader
from delia.utils.data_model import ImageAndSegmentationDataModel, PatientDataModel


class DefaultPatientDataFactory(BasePatientDataFactory):
    """
    Class that defines the methods that are used to get the patient data. The default factory consists in obtaining all
    the images.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            paths_to_segmentations: Optional[List[str]],
            series_descriptions: Optional[Dict[str, List[str]]],
            erase_unused_dicom_files: bool = False
    ):
        """
        Constructor of the class DefaultPatientDataFactory.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's image files.
        paths_to_segmentations : Optional[List[str]]
            List of paths to the patient's segmentation files.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        super().__init__(
            path_to_patient_folder=path_to_patient_folder,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions,
            erase_unused_dicom_files=erase_unused_dicom_files
        )

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        data = []
        for image_idx, image in enumerate(self._images_data):
            segmentations = []
            for path_to_segmentation in self._paths_to_segmentations:
                seg_header = DicomReader.get_dicom_header(path_to_dicom=path_to_segmentation)
                reference_uid = self.get_segmentation_reference_uid(seg_header)

                if image.dicom_header.SeriesInstanceUID == reference_uid:
                    segmentation_reader = SegmentationReader(
                        image=image,
                        path_to_segmentation=path_to_segmentation
                    )

                    segmentations.append(segmentation_reader.get_segmentation_data())

            if segmentations:
                image_and_segmentation_data = ImageAndSegmentationDataModel(
                    image=image,
                    segmentations=segmentations
                )
                data.append(image_and_segmentation_data)
            else:
                image_data = ImageAndSegmentationDataModel(image=image)
                data.append(image_data)

        patient_data = PatientDataModel(
            patient_id=self.patient_id,
            data=data
        )
        return patient_data


class SeriesDescriptionPatientDataFactory(BasePatientDataFactory):
    """
    Class that defines the methods that are used to get the patient data. The series description patient data factory
    consists in obtaining only the images that have the given series descriptions and their corresponding segmentations.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            paths_to_segmentations: Optional[List[str]],
            series_descriptions: Optional[Dict[str, List[str]]],
            erase_unused_dicom_files: bool = False
    ):
        """
        Constructor of the class SeriesDescriptionPatientDataFactory.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's image files.
        paths_to_segmentations : Optional[List[str]]
            List of paths to the patient's segmentation files.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        super().__init__(
            path_to_patient_folder=path_to_patient_folder,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions,
            erase_unused_dicom_files=erase_unused_dicom_files
        )

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        data = []
        for image_idx, image in enumerate(self._images_data):
            image_added = False
            series_description = image.dicom_header.SeriesDescription

            for series_key, list_of_series_descriptions in self._series_descriptions.items():
                if series_description in list_of_series_descriptions:
                    image.series_key = series_key

                    segmentations = []
                    for path_to_segmentation in self._paths_to_segmentations:
                        seg_header = DicomReader.get_dicom_header(path_to_dicom=path_to_segmentation)
                        reference_uid = self.get_segmentation_reference_uid(seg_header)

                        if image.dicom_header.SeriesInstanceUID == reference_uid:
                            segmentation_reader = SegmentationReader(
                                image=image,
                                path_to_segmentation=path_to_segmentation
                            )

                            segmentations.append(segmentation_reader.get_segmentation_data())

                    if segmentations:
                        image_and_segmentation_data = ImageAndSegmentationDataModel(
                            image=image,
                            segmentations=segmentations
                        )
                        data.append(image_and_segmentation_data)
                    else:
                        image_data = ImageAndSegmentationDataModel(image=image)
                        data.append(image_data)

                    image_added = True
                    break

            if image_added is False and self._erase_unused_dicom_files:
                self.erase_dicom_files(image)

        patient_data = PatientDataModel(
            patient_id=self.patient_id,
            data=data
        )
        return patient_data



"""
    @Title:             Database creation + handling of patients who failed.

    @Description:       Create the (possibly partial) HDF5 database, update the series description dictionary using the
                        information from the failed patients' images, and finally create the full HDF5 database.
"""

import env_examples  # Modifies path, DO NOT REMOVE

import json
from typing import Dict, List, Union

from delia.databases import PatientsDatabase
from delia.extractors import PatientsDataExtractor, PatientWhoFailed


def get_updated_series_descriptions(
        patients_who_failed: List[PatientWhoFailed],
        series_descriptions: Union[str, Dict[str, List[str]]]
) -> Union[str, Dict[str, List[str]]]:
    """
    Add a series description to the series description list of the given series key.

    Parameters
    ----------
    patients_who_failed : List[PatientWhoFailed]
        List of patients with one or more images not added to the HDF5 database due to the absence of the series in
        the patient record.
    series_descriptions : Union[str, Dict[str, List[str]]], default = None.
        A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
        the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
        series descriptions. The images associated with these series descriptions do not need to have a
        corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
        must be added to the database is to be able to add images without their segmentation.
    """
    for patient in patients_who_failed:
        for image, image_series_descriptions in patient.failed_images.items():
            if not any([series in series_descriptions[image] for series in patient.available_series_descriptions]):
                while True:
                    new_series_description = input(
                        f"\nNo available series for {image}. \nAvailable series are "
                        f"{patient.available_series_descriptions}. \nName of the series description to add : ")

                    print(f"Given series description name is {new_series_description}.")

                    if new_series_description in patient.available_series_descriptions:
                        print(f"Series description successfully added to the series descriptions json file.")
                        break
                    else:
                        print(f"Invalid series description! \n{new_series_description} not found in the patient's "
                              f"dicom files. Please try again.")

                series_descriptions[image] += [new_series_description]

    return series_descriptions


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #     Create database (some images of some patients might fail to be added to the database due to the         #
    #                         absence of the series descriptions in the patient record)                           #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        series_descriptions="data/incorrect_series_descriptions.json"
    )

    database = PatientsDatabase(path_to_database="data/patients_database.h5")

    patients_who_failed = database.create(
        patients_data_extractor=patients_data_extractor,
        overwrite_database=True
    )

    # Print list of patients who failed
    print(f"Patients who failed the pipeline : {patients_who_failed}")

    # ----------------------------------------------------------------------------------------------------------- #
    #                                          Update series descriptions                                         #
    # ----------------------------------------------------------------------------------------------------------- #
    with open("data/incorrect_series_descriptions.json", "r") as json_file:
        series_descriptions = json.load(json_file)

    updated_series_descriptions = get_updated_series_descriptions(
        patients_who_failed=patients_who_failed,
        series_descriptions=series_descriptions
    )

    with open("data/incorrect_series_descriptions.json", 'w', encoding='utf-8') as json_file:
        json.dump(updated_series_descriptions, json_file, ensure_ascii=False, indent=4)

    # ----------------------------------------------------------------------------------------------------------- #
    #                                           Create complete database                                          #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        series_descriptions=updated_series_descriptions
    )

    patients_who_failed = database.create(
        patients_data_extractor=patients_data_extractor,
        tags_to_use_as_attributes=[(0x0008, 0x103E), (0x0020, 0x000E), (0x0008, 0x0060)],
        overwrite_database=True
    )

    # Print list of patients who failed. This list should be empty now.
    print(f"Patients who failed the pipeline : {patients_who_failed}")



"""
    @file:              segmentation_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the SegmentationReader class which is used to read a given segmentation file
                        and transform its contents into the format of the SegmentationDataModel class.
"""

from .segmentation_context import SegmentationContext
from .factories.segmentation import Segmentation
from ...utils.data_model import ImageDataModel, SegmentationDataModel


class SegmentationReader:
    """
    A class used to read a given segmentation file and transform its contents into the standard format of the
    SegmentationDataModel class.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str,
    ):
        """
        Constructor of the class SegmentationReader.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        self._image = image
        self._path_to_segmentation = path_to_segmentation

    @property
    def __segmentation_context_manager(self) -> SegmentationContext:
        """
        Creates a SegmentationContext object.

        Returns
        -------
        segmentation_context : SegmentationContext
            Segmentation context manager.
        """
        return SegmentationContext(
            image=self._image,
            path_to_segmentation=self._path_to_segmentation
        )

    @property
    def __segmentation(self) -> Segmentation:
        """
        Creates a Segmentation object.

        Returns
        -------
        segmentation : Segmentation
            Segmentation.
        """
        return self.__segmentation_context_manager.create_segmentation()

    def get_segmentation_data(self) -> SegmentationDataModel:
        """
        Get the segmentation data from the path of the segmentation file.

        Returns
        -------
        segmentation_data : SegmentationDataModel
            A named tuple grouping the segmentation data.
        """
        segmentation_data = SegmentationDataModel(
            dicom_header=self.__segmentation_context_manager.segmentation_dicom_header,
            modality=self.__segmentation_context_manager.segmentation_strategy.modality,
            simple_itk_label_maps=self.__segmentation.simple_itk_label_maps
        )

        return segmentation_data



"""
    @file:              radiomics_dataset.py
    @Author:            Maxence Larose

    @Creation Date:     11/2022
    @Last modification: 11/2022

    @Description:       This file contains the RadiomicsDataset class which is used to create a dataset containing
                        radiomics information about images and segmentations.
"""

import logging
import os
from typing import Dict, List, Optional

import pandas as pd
from radiomics.featureextractor import RadiomicsFeatureExtractor

from delia.extractors import PatientsDataExtractor
from delia.readers.segmentation.segmentation_strategy import SegmentationStrategies

_logger = logging.getLogger(__name__)


class RadiomicsDataset:

    DEFAULT_INDEX_LABEL = "ID"

    def __init__(
            self,
            path_to_dataset: str
    ):
        """
        Used to interact with a dataset of radiomics features.

        Parameters
        ----------
        path_to_dataset : str
            Path to dataset.
        """
        self.path_to_dataset = path_to_dataset
        self._extractor = None

    @property
    def path_to_dataset(self) -> str:
        """
        Path to dataset.

        Returns
        -------
        path_to_dataset : str
            Path to dataset containing modality and organ names.
        """
        return self._path_to_dataset

    @path_to_dataset.setter
    def path_to_dataset(self, path_to_dataset: str) -> None:
        """
        Sets path to dataset.

        Parameters
        ----------
        path_to_dataset : str
            Path to dataset.
        """
        if path_to_dataset.endswith(".csv"):
            self._path_to_dataset = path_to_dataset
        else:
            self._path_to_dataset = f"{path_to_dataset}.csv"

    @property
    def extractor(self) -> Optional[RadiomicsFeatureExtractor]:
        """
        Radiomics extractor.

        Returns
        -------
        extractor : RadiomicsFeatureExtractor
            Radiomics extractor.
        """
        return self._extractor

    @extractor.setter
    def extractor(self, extractor: RadiomicsFeatureExtractor) -> None:
        """
        Set radiomics extractor.

        Parameters
        ----------
        extractor : RadiomicsFeatureExtractor
            Radiomics extractor.
        """
        self._extractor = extractor

    def _check_authorization_of_dataset_creation(
            self,
            overwrite_dataset: bool
    ) -> None:
        """
        Check if dataset's creation is allowed.

        Parameters
        ----------
        overwrite_dataset : bool
            Overwrite existing dataset.
        """
        if os.path.exists(self.path_to_dataset):
            if not overwrite_dataset:
                raise FileExistsError(
                    "The dataset already exists. You may overwrite it using overwrite_dataset = True."
                )
            else:
                _logger.info(f"Overwriting dataset with path : {self.path_to_dataset}.")
        else:
            _logger.info(f"Writing dataset with path : {self.path_to_dataset}.")

    @staticmethod
    def convert_list_of_dicts_to_dict_of_lists(list_of_dicts: List[Dict]) -> Dict[str, List]:
        """
        Converts a list of dictionaries into a dictionary of lists.

        Parameters
        ----------
        list_of_dicts : List[Dict]
            List of dictionaries. Example : [{"a": 1, "b": 2}, {"a": 3, "b": 4}].

        Returns
        -------
        dict_of_lists : Dict[str, List]
            Dictionary of lists. Example : {"a": [1, 3], "b": [2, 4]}.
        """
        dict_of_lists = {k: [dic[k] for dic in list_of_dicts] for k in list_of_dicts[0]}

        return dict_of_lists

    @staticmethod
    def convert_dict_of_lists_to_list_of_dicts(dict_of_lists: Dict[str, List]) -> List[Dict]:
        """
        Converts a dictionary of lists into a list of dictionaries.

        Parameters
        ----------
        dict_of_lists : Dict[str, List]
            Dictionary of lists. Example : {"a": [1, 3], "b": [2, 4]}.

        Returns
        -------
        dict_of_lists : Dict[str, List]
            List of dictionaries. Example : [{"a": 1, "b": 2}, {"a": 3, "b": 4}].
        """
        list_of_dicts = [dict(zip(dict_of_lists, t)) for t in zip(*dict_of_lists.values())]

        return list_of_dicts

    def create(
            self,
            patients_data_extractor: PatientsDataExtractor,
            organ: str,
            image_name: Optional[str] = None,
            image_modality: Optional[str] = None,
            segmentation_modality_to_prioritize: str = "SEG",
            overwrite_dataset: bool = False,
            **kwargs
    ) -> None:
        """
        Create radiomics datasets from all patients data.

        Parameters
        ----------
        patients_data_extractor : PatientsDataExtractor
            An object used to iterate on multiple patients' dicom files and segmentation files using the
            PatientDataReader to obtain all patients' data.
        organ : str
            The organ specifies the segment of the mask from which the radiomics features are to be acquired.
        image_name : Optional[str]
            Arbitrary name given to the image in the 'series_descriptions' dictionary. If 'series_descriptions' is
            None, use 'image_modality'.
        image_modality : Optional[str]
            Image modality. If 'series_descriptions' is NOT None, use 'image_name'.
        segmentation_modality_to_prioritize : str
            If there is multiple segmentation of the same organ for the same patient, this parameter sets the modality
            of the segmentation to prioritize, i.e. "SEG" or "RTSTRUCT". Default = "SEG".
        overwrite_dataset : bool
            Overwrite existing dataset, default = False.
        kwargs : {
            geometry_tolerance : float
                Image/Mask geometry mismatch tolerance.
            label : int, default = 1
                Integer, value of the label for which to extract features. If not specified, last specified label is
                used.
            label_channel : int, default = 0
                Integer, index of the channel to use when maskFilepath yields a SimpleITK.Image with a vector pixel
                type.
            voxel_based : bool, default = False
                If set to true, a voxel-based extraction is performed, segment-based otherwise.
        }
        """
        if not image_name and not image_modality:
            raise AssertionError(
                f"Parameters 'image_name' or 'image_modality' must be provided. Found 'image_name'={image_name} and "
                f"'image_modality'={image_modality}."
            )
        elif image_name and image_modality:
            raise AssertionError(
                f"Parameters 'image_name' OR 'image_modality' must be provided. Both are provided. Found "
                f"'image_name'={image_name} and 'image_modality'={image_modality}. One of the two must be None."
            )

        if segmentation_modality_to_prioritize not in SegmentationStrategies.get_available_modalities():
            raise AssertionError(
                f"Unknown 'segmentation_modality_to_prioritize={segmentation_modality_to_prioritize}'. Available "
                f"modalities are {SegmentationStrategies.get_available_modalities()}."
            )
        if not self.extractor:
            raise AssertionError("The 'extractor' needs to be set before creating the radiomics dataset.")

        self._check_authorization_of_dataset_creation(overwrite_dataset=overwrite_dataset)

        patients_ids: List[str] = []
        radiomics_features: List[Dict] = []
        for patient_idx, patient_dataset in enumerate(patients_data_extractor):
            patient_id = patient_dataset.patient_id
            patients_ids.append(patient_id)

            image, mask = None, None
            for image_idx, patient_image_data in enumerate(patient_dataset.data):
                series_key = patient_image_data.image.series_key
                modality = patient_image_data.image.dicom_header.Modality

                if image_name:
                    if series_key == image_name:
                        image = patient_image_data.image.simple_itk_image
                elif image_modality:
                    if modality == image_modality:
                        image = patient_image_data.image.simple_itk_image

                if image:
                    if patient_image_data.segmentations:
                        available_modalities = []
                        for seg in patient_image_data.segmentations:
                            if organ in seg.simple_itk_label_maps.keys():
                                available_modalities.append(seg.modality)
                        if segmentation_modality_to_prioritize in available_modalities:
                            seg_idx = available_modalities.index(segmentation_modality_to_prioritize)
                        else:
                            seg_idx = 0

                        segmentation = patient_image_data.segmentations[seg_idx]
                        for _organ, _mask in segmentation.simple_itk_label_maps.items():
                            if _organ == organ:
                                mask = _mask
                    break

            if not image:
                if image_name:
                    _logger.warning(f"No image found with name {image_name} for patient {patient_id}.")
                elif image_modality:
                    _logger.warning(f"No image found with modality {image_modality} for patient {patient_id}.")
            if not mask:
                _logger.warning(f"No mask found for organ {organ} for patient {patient_id}.")

            if image and mask:
                radiomics = self.extractor.execute(
                    imageFilepath=image,
                    maskFilepath=mask,
                    label=kwargs.get("label", None),
                    label_channel=kwargs.get("label_channel", None),
                    voxelBased=kwargs.get("voxel_based", False)
                )

                radiomics_features.append(radiomics)

        if radiomics_features:
            radiomics_features_of_multiple_patients = self.convert_list_of_dicts_to_dict_of_lists(
                list_of_dicts=radiomics_features
            )

            dataframe = pd.DataFrame(
                data=radiomics_features_of_multiple_patients,
                index=patients_ids
            )

            dataframe.to_csv(
                path_or_buf=self.path_to_dataset,
                sep=",",
                index=True,
                index_label=self.DEFAULT_INDEX_LABEL
            )

        else:
            _logger.error(
                f"No images found for all patients. The radiomics dataset with path {self.path_to_dataset} was "
                f"therefore not created."
            )

        patients_data_extractor.reset()
        patients_data_extractor.close()

    def read(self) -> pd.DataFrame:
        """
        Read dataset.

        Returns
        -------
        dataframe : pd.DataFrame
            Radiomics pandas dataframe.
        """
        dataframe = pd.read_csv(filepath_or_buffer=self.path_to_dataset, index_col=self.DEFAULT_INDEX_LABEL)

        return dataframe



"""
    @file:              applications.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 11/2022

    @Description:       This file contains the apply_transforms function which is used to apply transformations to
                        images and segmentations.
"""

from copy import deepcopy
from typing import Dict, Hashable, NamedTuple, Union

from monai.transforms import apply_transform as monai_apply_transform
from monai.transforms import Compose, EnsureChannelFirstD
from monai.transforms import MapTransform as MonaiMapTransform
import SimpleITK as sitk

from delia.transforms.array_space.transform import ArraySpaceTransform
from delia.transforms.physical_space.transform import PhysicalSpaceTransform, ImageData
from delia.transforms.data.transform import DataTransform
from delia.transforms.tools import convert_to_numpy, Mode, set_transforms_keys
from delia.utils.data_model import PatientDataModel


class _SitkImageInfo(NamedTuple):
    spacing: tuple
    origin: tuple
    direction: tuple


def apply_transforms(
        patient_dataset: PatientDataModel,
        transforms: Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
) -> None:
    """
    Applies transforms on images and segmentations.

    Parameters
    ----------
    patient_dataset : PatientDataModel
        A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
        segmentation data extracted from the segmentation files.
    transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
        A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on the
        SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that represents
        the image. DataTransform transforms the data using other a patient's other images or segmentations. The keys
        for images are assumed to be the arbitrary series key set in 'series_descriptions'. For segmentation, keys are
        organ names. Note that if 'series_descriptions' is None, the keys for images are assumed to be modalities.
    """
    set_transforms_keys(patient_dataset=patient_dataset)

    if isinstance(transforms, Compose):
        for t in transforms.transforms:
            if isinstance(t, DataTransform):
                _apply_data_transform(
                    transform=t,
                    patient_dataset=patient_dataset
                )
            elif isinstance(t, (PhysicalSpaceTransform, MonaiMapTransform)):
                _apply_transform_on_segmentations(
                    transform=t,
                    patient_dataset=patient_dataset
                )
                _apply_transform_on_images(
                    transform=t,
                    patient_dataset=patient_dataset
                )
    else:
        if isinstance(transforms, DataTransform):
            _apply_data_transform(
                transform=transforms,
                patient_dataset=patient_dataset
            )
        elif isinstance(transforms, (PhysicalSpaceTransform, MonaiMapTransform)):
            _apply_transform_on_segmentations(
                transform=transforms,
                patient_dataset=patient_dataset
            )
            _apply_transform_on_images(
                transform=transforms,
                patient_dataset=patient_dataset
            )


def _apply_data_transform(
        patient_dataset: PatientDataModel,
        transform: DataTransform
) -> None:
    """
    Applies single transform on patient dataset.

    Parameters
    ----------
    patient_dataset : PatientDataModel
        A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
        segmentation data extracted from the segmentation files.
    transform : DataTransform
        A data transformation to apply.
    """
    data = {data.image.transforms_key: data for data in patient_dataset.data}

    transformed_data = monai_apply_transform(transform=transform, data=data)

    patient_dataset.data = list(transformed_data.values())


def _apply_transform_on_images(
        patient_dataset: PatientDataModel,
        transform: Union[MonaiMapTransform, PhysicalSpaceTransform]
) -> None:
    """
    Applies single transform on images.

    Parameters
    ----------
    patient_dataset : PatientDataModel
        A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
        segmentation data extracted from the segmentation files.
    transform : Union[MonaiMapTransform, PhysicalSpaceTransform]
        A transformation to apply on images. PhysicalSpaceTransform are applied in the physical space, i.e on the
        SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that represents
        the image. The keys for images are assumed to be the arbitrary series key set in 'series_descriptions'. For
        segmentation, keys are organ names. Note that if 'series_descriptions' is None, the keys for images are
        assumed to be modalities.
    """
    images = {
        data.image.transforms_key: ImageData(
            simple_itk_image=data.image.simple_itk_image,
            dicom_header=data.image.dicom_header
        )
        for data in patient_dataset.data
    }

    transformed_images_dict = _apply_transform(transform=transform, data=images, mode=Mode.IMAGE)
    for data in patient_dataset.data:
        image = data.image
        image.simple_itk_image = transformed_images_dict[image.transforms_key]


def _apply_transform_on_segmentations(
        patient_dataset: PatientDataModel,
        transform: Union[MonaiMapTransform, PhysicalSpaceTransform]
) -> None:
    """
    Applies single transform on segmentations.

    Parameters
    ----------
    patient_dataset : PatientDataModel
        A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
        segmentation data extracted from the segmentation files.
    transform : Union[MonaiMapTransform, PhysicalSpaceTransform]
        A transformation to apply on segmentations. PhysicalSpaceTransform are applied in the physical space, i.e on the
        SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that represents
        the image. Image keys are assumed to be arbitrary series keys defined in 'series_descriptions'. For the
        label maps, the keys are organ names. Note that if 'series_descriptions' is None, the image keys are
        assumed to be modality names.
    """
    images = {
        data.image.transforms_key: ImageData(
            simple_itk_image=data.image.simple_itk_image,
            dicom_header=data.image.dicom_header
        )
        for data in patient_dataset.data
    }

    for image_and_segmentation_data in patient_dataset.data:
        segmentations = image_and_segmentation_data.segmentations

        if segmentations:
            for segmentation_data in segmentations:
                temp_dict = deepcopy(images)
                for organ_name, label_map in segmentation_data.simple_itk_label_maps.items():
                    temp_dict[organ_name] = ImageData(simple_itk_image=label_map)

                transformed_dict = _apply_transform(transform=transform, data=temp_dict, mode=Mode.SEGMENTATION)

                for img_key in images.keys():
                    transformed_dict.pop(img_key)

                segmentation_data.simple_itk_label_maps = transformed_dict


def _apply_transform(
        data: Dict[str, ImageData],
        transform: Union[MonaiMapTransform, PhysicalSpaceTransform],
        mode: Mode
) -> Dict[Hashable, sitk.Image]:
    """
    Applies single transform on segmentations.

    Parameters
    ----------
    data : Dict[str, ImageData]
        A Python dictionary that contains ImageData to be transformed.
    transform : Union[MonaiMapTransform, PhysicalSpaceTransform]
        A transformation to apply. PhysicalSpaceTransform are applied in the physical space, i.e on the SimpleITK image,
        while MonaiMapTransform are applied in the array space, i.e on the numpy array that represents the image. The
        keys for images are assumed to be the arbitrary series key set in 'series_descriptions'. For segmentation,
        keys are organ names. Note that if 'series_descriptions' is None, the keys for images are assumed to be
        modalities.
    mode : Mode
        The mode.

    Returns
    -------
    transformed_data : Dict[Hashable, sitk.Image]
        A dictionary of transformed SimpleITK images.
    """
    if isinstance(transform, PhysicalSpaceTransform):
        return _apply_physical_transform(transform=transform, data=data, mode=mode)
    elif isinstance(transform, ArraySpaceTransform):
        data = {k: v.simple_itk_image for k, v in data.items()}
        return _apply_array_transform(transform=transform, data=data, mode=mode)
    elif isinstance(transform, MonaiMapTransform):
        data = {k: v.simple_itk_image for k, v in data.items()}
        return _apply_monai_transforms(transform=transform, data=data)


def _apply_physical_transform(
        data: Dict[str, ImageData],
        transform: PhysicalSpaceTransform,
        mode: Mode
) -> Dict[Hashable, sitk.Image]:
    """
    Apply a PhysicalSpaceTransform.

    Parameters
    ----------
    data : Dict[str, ImageData]
        A Python dictionary that contains ImageData to be transformed.
    transform : PhysicalSpaceTransform
        A transformation to apply to images and segmentations in the physical space, i.e on the SimpleITK image. Keys
        are assumed to be modality names for images and organ names for segmentations.
    mode : Mode
        Mode.

    Returns
    -------
    transformed_data : Dict[Hashable, sitk.Image]
        A dictionary of transformed SimpleITK images.
    """
    transform.mode = mode.value

    transformed_data = monai_apply_transform(transform=transform, data=data)
    transform.mode = Mode.NONE.value

    for k, v in transformed_data.items():
        if isinstance(v, ImageData):
            transformed_data[k] = v.simple_itk_image

    return transformed_data


def _apply_array_transform(
        data: Dict[str, sitk.Image],
        transform: ArraySpaceTransform,
        mode: Mode
) -> Dict[Hashable, sitk.Image]:
    """
    Apply a ArraySpaceTransform.

    Parameters
    ----------
    data : Dict[str, sitk.Image]
        A Python dictionary that contains ImageData to be transformed.
    transform : ArraySpaceTransform
        A transformation to apply to images and segmentations in the physical space, i.e on the SimpleITK image. Keys
        are assumed to be modality names for images and organ names for segmentations.
    mode : Mode
        Mode.

    Returns
    -------
    transformed_data : Dict[Hashable, sitk.Image]
        A dictionary of transformed SimpleITK images.
    """
    transform.mode = mode.value
    transformed_data = _apply_monai_transforms(data=data, transform=transform)
    transform.mode = Mode.NONE.value

    return transformed_data


def _apply_monai_transforms(
        data: Dict[str, sitk.Image],
        transform: MonaiMapTransform
) -> Dict[Hashable, sitk.Image]:
    """
    Apply a MonaiMapTransform.

    Parameters
    ----------
    data : Dict[str, sitk.Image]
        A dictionary of SimpleITK images to be transformed.
    transform : MonaiMapTransform
        A transformation to apply to images and segmentations in the array space, i.e on the numpy array that
        represents the image. Keys are assumed to be modality names for images and organ names for segmentations.

    Returns
    -------
    transformed_data : Dict[Hashable, sitk.Image]
        A dictionary of transformed SimpleITK images.
    """
    ensure_channel_first_d = EnsureChannelFirstD(keys=transform.keys, allow_missing_keys=True)
    transform = Compose([ensure_channel_first_d, transform])

    info = {}
    temp_dict = {}
    for k, img in data.items():
        info[k] = _SitkImageInfo(spacing=img.GetSpacing(), origin=img.GetOrigin(), direction=img.GetDirection())
        temp_dict[k] = sitk.GetArrayFromImage(data[k])

    transformed_data = monai_apply_transform(transform=transform, data=temp_dict)

    for k, img in transformed_data.items():
        transformed_img_array = convert_to_numpy(transformed_data[k])

        transformed_img_sitk = sitk.GetImageFromArray(transformed_img_array)
        transformed_img_sitk.SetSpacing(info[k].spacing)
        transformed_img_sitk.SetOrigin(info[k].origin)
        transformed_img_sitk.SetDirection(info[k].direction)

        transformed_data[k] = transformed_img_sitk

    return transformed_data



"""
    @file:              patient_data_query_context.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 03/2022

    @Description:       This file contains the class PatientDataQueryContext that is used as a context class where
                        strategies are types of requests the client could ask the PatientDataReader class.
"""

from typing import Dict, List, Optional

from delia.utils.data_model import PatientDataModel
from .patient_data_query_strategy import PatientDataQueryStrategy, PatientDataQueryStrategies


class PatientDataQueryContext:
    """
    A class used as a context class where strategies are types of requests the client could make to the
    PatientDataReader class to get the patient's data.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            paths_to_segmentations: Optional[List[str]],
            series_descriptions: Dict[str, List[str]],
            erase_unused_dicom_files: bool = False
    ):
        """
        Constructor of the PatientDataQueryContext class.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's DICOM files.
        paths_to_segmentations : Optional[List[str]]
            List of paths to the patient's segmentation files.
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        self._path_to_patient_folder = path_to_patient_folder
        self._paths_to_segmentations = paths_to_segmentations
        self._series_descriptions = series_descriptions
        self._erase_unused_dicom_files = erase_unused_dicom_files

    @property
    def patient_data_query_strategy(self) -> PatientDataQueryStrategy:
        """
        Patient data query strategy corresponding to the given series descriptions configuration.

        Returns
        -------
        patient_data_query_strategy : PatientDataQueryStrategy
            Patient data query strategy.
        """
        if self._series_descriptions:
            return PatientDataQueryStrategies.SERIES_DESCRIPTION.value
        else:
            return PatientDataQueryStrategies.DEFAULT.value

    @property
    def _patient_data_factory_instance(self) -> PatientDataQueryStrategy.factory:
        """
        The factory class instance corresponding to the class of the given patient data strategy.

        Returns
        -------
        _patient_data_factory_instance : PatientDataQueryStrategy.factory
            Factory class instance used to get a patient's data.
        """
        patient_data_factory_instance = self.patient_data_query_strategy.factory(
            path_to_patient_folder=self._path_to_patient_folder,
            paths_to_segmentations=self._paths_to_segmentations,
            series_descriptions=self._series_descriptions,
            erase_unused_dicom_files=self._erase_unused_dicom_files
        )

        return patient_data_factory_instance

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        return self._patient_data_factory_instance.create_patient_data()



"""
    @file:              segmentation_context.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the class SegmentationContext that is used as a context class where
                        strategies are types of ways to load the segmentation data, or more precisely, types of
                        segmentation object factory.
"""

import pydicom

from .segmentation_strategy import SegmentationStrategy, SegmentationStrategies
from .factories.segmentation import Segmentation
from ...utils.data_model import ImageDataModel


class SegmentationContext:
    """
    A class used as a context class where strategies are types of ways to load the segmentation data, or more precisely,
    types of segmentation object factory. Strategies are entirely defined by the extension of the given file and so, by
    the path of the segmentation.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str
    ):
        """
        Constructor of the SegmentationContext class.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        self._image = image
        self._path_to_segmentation = path_to_segmentation

    @property
    def path_to_segmentation(self) -> str:
        """
        Path to segmentation property.

        Returns
        -------
        path_to_segmentation : str
            The path to the segmentation file.
        """
        return self._path_to_segmentation

    @path_to_segmentation.setter
    def path_to_segmentation(self, path_to_segmentation: str) -> None:
        """
        Path to segmentation setter.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        """
        self._path_to_segmentation = path_to_segmentation

    @property
    def segmentation_modality(self) -> str:
        """
        Segmentation modality.

        Returns
        -------
        segmentation_modality : str
            Segmentation modality.
        """
        modality = self.segmentation_dicom_header.Modality
        available_modalities = SegmentationStrategies.get_available_modalities()

        assert modality in available_modalities, (
            f"The given segmentation file ({self.path_to_segmentation}) is of modality {modality}. However, the "
            f"available modalities for the segmentations are {available_modalities}."
        )

        return modality

    @property
    def segmentation_dicom_header(self) -> pydicom.dataset.FileDataset:
        """
        Segmentation dicom header.

        Returns
        -------
        segmentation_dicom_header : pydicom.dataset.FileDataset
            Segmentation dicom header.
        """
        header = pydicom.dcmread(self.path_to_segmentation, stop_before_pixels=True)

        return header

    @property
    def segmentation_strategy(self) -> SegmentationStrategy:
        """
        Segmentation strategy corresponding to the given segmentation modality.

        Returns
        -------
        segmentation_strategy : SegmentationStrategy
            Segmentation strategy.
        """
        modality = self.segmentation_modality
        for segmentation_category in list(SegmentationStrategies):
            if modality == segmentation_category.value.modality:
                return segmentation_category.value

    @property
    def _segmentation_factory_instance(self) -> SegmentationStrategy.factory:
        """
        The segmentation factory instance corresponding to the class of the given segmentation category.

        Returns
        -------
        _segmentation_factory_instance : SegmentationStrategy.factory
            Factory class instance used to get the label maps and the segmentation metadata from a segmentation file.
        """
        return self.segmentation_strategy.factory(
            image=self._image,
            path_to_segmentation=self.path_to_segmentation
        )

    def create_segmentation(self) -> Segmentation:
        """
        Creates a Segmentation object.

        Returns
        -------
        segmentation : Segmentation
            Segmentation.
        """
        return self._segmentation_factory_instance.create_segmentation()



"""
    @file:              pet_to_suv.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 11/2022

    @Description:       This file contains the PETtoSUVd transform.
"""

from datetime import datetime
import logging
from typing import Dict, Hashable

import numpy as np
import pydicom
import SimpleITK as sitk

from delia.transforms.physical_space.transform import PhysicalSpaceTransform, ImageData, KeysCollection, Mode

_logger = logging.getLogger(__name__)


class DefaultParams:
    WEIGHT = 75000  # [grams]
    SCAN_INJECTION_DELAY = 1.75 * 3600  # [seconds]
    HALF_LIFE = 6588  # [seconds]
    TOTAL_INJECTED_DOSE = 420_000_000  # [becquerels]


class PETtoSUVd(PhysicalSpaceTransform):
    """
    Transform a PET image to Standardized uptake value (SUV) in g/ml..
    """

    PET_MODALITY_NAME = "PT"

    def __init__(
            self,
            keys: KeysCollection
    ):
        """
        Initialize transform.

        Parameters
        ----------
        keys : KeysCollection
            Keys of the corresponding items to be transformed. Image keys are assumed to be arbitrary series keys
            defined in 'series_descriptions'. For the label maps, the keys are organ names. Note that if
            'series_descriptions' is None, the image keys are assumed to be modality names.
        """
        super().__init__(keys=keys)

    @staticmethod
    def get_patient_weight(header: pydicom.dataset.FileDataset) -> float:
        """
        Get patient weight or estimate it.

        Parameters
        ----------
        header : pydicom.dataset.FileDataset
            Dicom header.

        Returns
        -------
        weight : float
            Patient weight in grams.
        """
        if hasattr(header, "PatientWeight"):
            return float(header.PatientWeight) * 1000
        else:
            _logger.warning(
                f"Attribute 'PatientWeight' doesn't exist. Using estimated patient weight of "
                f"{DefaultParams.WEIGHT / 1000} kg."
            )

            return DefaultParams.WEIGHT

    @staticmethod
    def get_time_delay_between_injection_and_scan(header: pydicom.dataset.FileDataset) -> float:
        """
        Get time delay between radiopharmaceutical injection and scan.

        Parameters
        ----------
        header : pydicom.dataset.FileDataset
            Dicom header.

        Returns
        -------
        time_delay : float
            Time delay in seconds.
        """
        if hasattr(header, "AcquisitionTime"):
            try:
                scan_time = datetime.strptime(header.AcquisitionTime, '%H%M%S.%f')

                if hasattr(header.RadiopharmaceuticalInformationSequence[0], "RadiopharmaceuticalStartTime"):
                    injection_time = datetime.strptime(
                        header.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime, '%H%M%S.%f'
                    )

                    return (scan_time - injection_time).seconds
                else:
                    _logger.warning(
                        f"Attribute 'RadiopharmaceuticalStartTime' doesn't exist. Using estimated time delay between "
                        f"injection and scan of {DefaultParams.SCAN_INJECTION_DELAY / 60} minutes, i.e. 90 min waiting "
                        f"time + 15 min preparation."
                    )

                    return DefaultParams.SCAN_INJECTION_DELAY
            except ValueError:
                _logger.warning(
                    f"Unknown time data format (Expected '%H%M%S.%f). Using estimated time delay between injection "
                    f"and scan of {DefaultParams.SCAN_INJECTION_DELAY / 60} minutes, i.e. 90 min waiting time + 15 min "
                    f"preparation."
                )
                return DefaultParams.SCAN_INJECTION_DELAY
        else:
            _logger.warning(
                f"Attribute 'AcquisitionTime' doesn't exist. Using estimated time delay between injection and "
                f"scan of {DefaultParams.SCAN_INJECTION_DELAY / 60} minutes, i.e. 90 min waiting time + 15 min "
                f"preparation."
            )

            return DefaultParams.SCAN_INJECTION_DELAY

    @staticmethod
    def get_radionuclide_half_life(header: pydicom.dataset.FileDataset) -> float:
        """
        Get radionuclide half-life.

        Parameters
        ----------
        header : pydicom.dataset.FileDataset
            Dicom header.

        Returns
        -------
        half_life : float
            Half life in seconds.
        """
        if hasattr(header.RadiopharmaceuticalInformationSequence[0], "RadionuclideHalfLife"):
            return float(header.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife)
        else:
            _logger.warning(
                f"Attribute 'RadionuclideHalfLife' doesn't exist. Using estimated radionuclide half-life of "
                f"{DefaultParams.HALF_LIFE} seconds."
            )

            return DefaultParams.HALF_LIFE

    @staticmethod
    def get_radionuclide_total_injected_dose(header: pydicom.dataset.FileDataset) -> float:
        """
        Get radionuclide total injected dose.

        Parameters
        ----------
        header : pydicom.dataset.FileDataset
            Dicom header.

        Returns
        -------
        injected_dose : float
            Injected dose in Bq.
        """
        if hasattr(header.RadiopharmaceuticalInformationSequence[0], "RadionuclideTotalDose"):
            return float(header.RadiopharmaceuticalInformationSequence[0].RadionuclideTotalDose)
        else:
            _logger.warning(
                f"Attribute 'RadionuclideTotalDose' doesn't exist. Using estimated total injected dose of "
                f"{DefaultParams.TOTAL_INJECTED_DOSE / 1e6} MBq."
            )

            return DefaultParams.TOTAL_INJECTED_DOSE

    def compute_suv(self, sitk_pet_image: sitk.Image, dicom_header: pydicom.dataset.FileDataset) -> sitk.Image:
        """
        Computes Standardized uptake value (SUV) in g/ml.

        Parameters
        ----------
        sitk_pet_image : sitk.Image
            SimpleITK image of a PET scan where pixel intensities are related to FDG uptake.
        dicom_header : pydicom.dataset.FileDataset
            Dicom header.

        Returns
        -------
        sitk_suv_image : sitk.Image
            SimpleITK image of a PET scan where pixel intensities are SUV.
        """
        weight_grams = self.get_patient_weight(dicom_header)

        if hasattr(dicom_header, "RadiopharmaceuticalInformationSequence"):
            # Time delay between injection and acquisition
            time_delay = self.get_time_delay_between_injection_and_scan(header=dicom_header)

            # Half Life for Radionuclide # seconds
            half_life = self.get_radionuclide_half_life(header=dicom_header)

            # Total dose injected for Radionuclide
            injected_dose = self.get_radionuclide_total_injected_dose(header=dicom_header)

            # Calculate decay
            decay = np.exp(-np.log(2) * time_delay / half_life)

            # Calculate the dose decayed during procedure
            injected_dose_decay = injected_dose * decay  # [Bq]
        else:
            _logger.warning(
                f"Attribute 'RadiopharmaceuticalInformationSequence' doesn't exist. Using estimated time delay "
                f"between injection and scan of {DefaultParams.SCAN_INJECTION_DELAY / 60} minutes (90 min "
                f"waiting time, 15 min preparation), estimated radionuclide half-life of "
                f"{DefaultParams.HALF_LIFE} seconds and estimated total injected dose of "
                f"{DefaultParams.TOTAL_INJECTED_DOSE / 1e6} MBq."
            )

            decay = np.exp(-np.log(2) * DefaultParams.SCAN_INJECTION_DELAY / DefaultParams.HALF_LIFE)
            injected_dose_decay = DefaultParams.TOTAL_INJECTED_DOSE * decay

        # Calculate SUV [g/ml]
        suv = sitk_pet_image * weight_grams / injected_dose_decay

        return suv

    def __call__(self, data: Dict[str, ImageData]) -> Dict[Hashable, sitk.Image]:
        """
        Resample an itk_image to new out_spacing.

        Parameters
        ----------
        data : Dict[str, ImageData]
            A Python dictionary that contains ImageData.

        Returns
        -------
        transformed_image : Dict[Hashable, sitk.Image]
            A Python dictionary that contains transformed SimpleITK images.
        """
        d = dict(data)

        for key in self.key_iterator(d):
            if self._mode == Mode.NONE:
                raise AssertionError("Transform mode must be set before __call__.")
            elif not d[key].dicom_header:
                raise AssertionError(
                    f"The 'PETtoSUV' can't be used on segmentations. The 'PETtoSUV' can only be used on "
                    f"{self.PET_MODALITY_NAME} modality."
                )
            elif str(d[key].dicom_header.Modality) != self.PET_MODALITY_NAME:
                raise AssertionError(
                    f"Tried using 'PETtoSUV' transform on {d[key].dicom_header.Modality} modality. The 'PETtoSUV' "
                    f"can only be used on {self.PET_MODALITY_NAME} modality."
                )
            else:
                d[key] = self.compute_suv(sitk_pet_image=d[key].simple_itk_image, dicom_header=d[key].dicom_header)

        return d


PETtoSUVD = PETtoSUVDict = PETtoSUVd



"""
    @Title:             Logging + database creation.

    @Description:       Configure logging and create database.
"""

import env_examples  # Modifies path, DO NOT REMOVE

from delia.databases import PatientsDatabase
from delia.extractors import PatientsDataExtractor
from delia.transforms import (
    MatchingResampleD,
    PETtoSUVD,
    ResampleD
)
from monai.transforms import (
    CenterSpatialCropD,
    Compose,
    ScaleIntensityRangeD
)


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #                                         Logs Setup (Optional)                                               #
    # ----------------------------------------------------------------------------------------------------------- #
    env_examples.configure_logging("logging_conf.yaml")

    # ----------------------------------------------------------------------------------------------------------- #
    #                                      Create patients data extractor                                         #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        series_descriptions="data/series_descriptions.json",
        transforms=Compose(
            [
                ResampleD(keys=["CT_THORAX", "PET", "Heart"], out_spacing=(1.5, 1.5, 1.5)),
                MatchingResampleD(reference_image_key="CT_THORAX", matching_keys=["PET", "Heart"]),
                CenterSpatialCropD(keys=["CT_THORAX", "PET", "Heart"], roi_size=(1000, 160, 160)),
                ScaleIntensityRangeD(keys=["CT_THORAX"], a_min=-250, a_max=500, b_min=0, b_max=1, clip=True),
                PETtoSUVD(keys=["PET"])
            ]
        )
    )

    # ----------------------------------------------------------------------------------------------------------- #
    #                                                Create database                                              #
    # ----------------------------------------------------------------------------------------------------------- #
    database = PatientsDatabase(path_to_database="data/patients_database.h5")

    database.create(
        patients_data_extractor=patients_data_extractor,
        tags_to_use_as_attributes=[(0x0008, 0x103E), (0x0020, 0x000E), (0x0008, 0x0060)],
        overwrite_database=True
    )

    print("Database length:", len(database))
    print("R01-005 patient :", database["R01-005"])
    print("R01-005 patient :", database[0])



"""
    @file:              matching_resample.py
    @Author:            Maxence Larose

    @Creation Date:     03/2022
    @Last modification: 03/2022

    @Description:       This file contains the MatchingResampled transform.
"""

from typing import Callable, Dict, Hashable

import SimpleITK as sitk

from delia.transforms.physical_space.transform import PhysicalSpaceTransform, ImageData, KeysCollection, Mode


class MatchingResampled(PhysicalSpaceTransform):
    """
    Resample matching images to the spacing, size, origin and direction of a given reference image.
    """

    def __init__(
            self,
            reference_image_key: str,
            matching_keys: KeysCollection,
            interpolator: Callable = sitk.sitkBSpline,
    ):
        """
        Initializes images.

        Parameters
        ----------
        reference_image_key : str
            Key of the image from which the spacing, size, origin and direction are taken.
        matching_keys : KeysCollection
            Keys of the corresponding items to be transformed using the calculated coordinates of spatial bounding box
            for foreground on the image. Image keys are assumed to be arbitrary series keys defined in
            'series_descriptions'. For the label maps, the keys are organ names. Note that if 'series_descriptions' is
            None, the image keys are assumed to be modality names.
        interpolator : Callable
            The interpolator to be used for resampling the images. Default = sitk.sitkBSpline.
        """
        keys = [key for key in matching_keys] + [reference_image_key]
        super().__init__(keys=keys)
        self._reference_image_key = reference_image_key
        self._interpolator = interpolator

    def __call__(self, data: Dict[str, ImageData]) -> Dict[Hashable, sitk.Image]:
        """
        Resamples matching itk images to the spacing, size, origin and direction of a given reference image.

        Parameters
        ----------
        data : Dict[str, ImageData]
            A Python dictionary that contains ImageData.

        Returns
        -------
        transformed_image : Dict[Hashable, sitk.Image]
            A Python dictionary that contains transformed SimpleITK images.
        """
        d = dict(data)

        reference_image = None
        for key in self.key_iterator(d):
            if key == self._reference_image_key:
                reference_image = d[key].simple_itk_image

        for key in self.key_iterator(d):
            if key != self._reference_image_key:

                if self._mode == Mode.NONE:
                    raise AssertionError("Transform mode must be set before __call__.")
                elif self._mode == Mode.IMAGE:
                    interpolator = self._interpolator
                elif self._mode == Mode.SEGMENTATION:
                    interpolator = sitk.sitkNearestNeighbor
                else:
                    raise ValueError("Unknown transform mode.")

                d[key] = sitk.Resample(
                    image1=d[key].simple_itk_image,
                    referenceImage=reference_image,
                    interpolator=interpolator
                )

        return d


MatchingResampleD = MatchingResampleDict = MatchingResampled



"""
    @file:              transforms_history.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 10/2022

    @Description:       This file contains the TransformsHistory class which is used to define a simple transforms
                        history container.
"""


from typing import Optional, Union

from monai.transforms import Compose, MapTransform


class TransformsHistory:
    """
    A simple history of the transforms applied on an image.
    """

    def __init__(self, transforms: Optional[Union[Compose, MapTransform]] = None):
        """
        Initialize history.

        Parameters
        ----------
        transforms : Optional[Union[Compose, MapTransform]]
            Transforms.
        """
        self.history = []

        if transforms:
            self.append(transforms)

    def append(self, transforms: Union[Compose, MapTransform]) -> None:
        """
        Append transforms to history.

        Parameters
        ----------
        transforms : Optional[Union[Compose, MapTransform]]
            Transforms.
        """
        if isinstance(transforms, MapTransform):
            self.history.append(transforms.__dict__)
        elif isinstance(transforms, Compose):
            for transform in transforms.__dict__["transforms"]:
                self.history.append(dict({"name": transform.__class__.__name__}, **vars(transform)))
        else:
            raise AssertionError("'transforms' must be Union[Compose, MapTransform].")

    @staticmethod
    def serialize(transforms: Union[Compose, MapTransform]) -> Union[dict, str]:
        """
        Serialize given transforms.

        Parameters
        ----------
        transforms : Optional[Union[Compose, MapTransform]]
            Transforms.

        Returns
        -------
        serialized_transforms : Union[dict, str]
            Serialized transforms.
        """
        try:
            return vars(transforms)
        except (AttributeError, TypeError):
            return "<not serializable>"



"""
    @file:              matching_crop_foreground.py
    @Author:            Maxence Larose

    @Creation Date:     11/2022
    @Last modification: 11/2022

    @Description:       This file contains the MatchingCropForeground transform.
"""

from typing import Collection, Dict, Hashable, Union

from monai.transforms import CropForeground, SpatialCrop
import numpy as np

from .transform import ArraySpaceTransform

KeysCollection = Union[Collection[Hashable], Hashable]


class MatchingCropForegroundd(ArraySpaceTransform):
    """
    Performs CropForeground on an image, get the used coordinates of spatial bounding box for foreground and apply this
    crop on other matching images.
    """

    def __init__(
            self,
            reference_image_key: str,
            matching_keys: KeysCollection,
    ):
        """
        Initialize output spacing.

        Parameters
        ----------
        reference_image_key : str
            Key of the image from which to crop foreground and to get coordinates of spatial bounding box for
            foreground. Image keys are assumed to be arbitrary series keys defined in 'series_descriptions'. Note that
            if 'series_descriptions' is None, the image keys are assumed to be modality names.
        matching_keys : KeysCollection
            Keys of the corresponding items to be transformed using the calculated coordinates of spatial bounding box
            for foreground on the image. Image keys are assumed to be arbitrary series keys defined in
            'series_descriptions'. For the label maps, the keys are organ names. Note that if 'series_descriptions' is
            None, the image keys are assumed to be modality names.
        """
        keys = [key for key in matching_keys] + [reference_image_key]
        super().__init__(keys=keys)

        self._reference_image_key = reference_image_key
        self._crop_foreground = CropForeground(return_coords=True)

    def __call__(self, data: Dict[Hashable, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Performs CropForeground on an image, get the used coordinates of spatial bounding box for foreground and apply
        this crop on other matching images.

        Parameters
        ----------
        data : Dict[Hashable, np.ndarray]
            A Python dictionary that contains numpy ndarray.

        Returns
        -------
        transformed_data : Dict[Hashable, np.ndarray]
            A Python dictionary that contains transformed numpy ndarray.
        """
        d = dict(data)

        start, end = None, None
        for key in self.key_iterator(d):
            if key == self._reference_image_key:
                d[key], start, end = self._crop_foreground(d[key])

        spatial_crop = SpatialCrop(roi_start=start, roi_end=end)

        for key in self.key_iterator(d):
            if key != self._reference_image_key:
                d[key] = spatial_crop(d[key])

        return d


MatchingCropForegroundD = MatchingCropForegroundDict = MatchingCropForegroundd



from .matching_centroid_spatial_crop import (
    MatchingCentroidSpatialCropd,
    MatchingCentroidSpatialCropD,
    MatchingCentroidSpatialCropDict
)
from .matching_crop_foreground import (
    MatchingCropForegroundd,
    MatchingCropForegroundD,
    MatchingCropForegroundDict
)



"""
    @file:              patients_data_generator.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains the PatientsDataGenerator class which is used to iterate on multiple
                        patients' dicom files and segmentation files using the PatientDataReader to obtain all patients'
                        data. The PatientsDataGenerator inherits from the Generator abstract class.
"""

from collections.abc import Generator
from copy import deepcopy
import json
import logging
import os
from typing import Dict, List, NamedTuple, Optional, Union

from monai.transforms import Compose
from monai.transforms import MapTransform as MonaiMapTransform

from delia.readers.patient_data.patient_data_reader import PatientDataReader
from delia.transforms.data.transform import DataTransform
from delia.transforms.physical_space.transform import PhysicalSpaceTransform
from delia.utils.data_model import PatientDataModel

_logger = logging.getLogger(__name__)


class PatientWhoFailed(NamedTuple):
    id: str
    failed_images: Dict[str, List[str]]
    available_series_descriptions: List[str]


class PatientsDataExtractor(Generator):
    """
    A class used to iterate on multiple patients' dicom files and segmentation files using the PatientDataReader to
    obtain all patients' data. The PatientsDataGenerator inherits from the Generator abstract class.
    """

    def __init__(
            self,
            path_to_patients_folder: str,
            series_descriptions: Optional[Union[str, Dict[str, List[str]]]] = None,
            transforms: Optional[Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]] = None,
            erase_unused_dicom_files: bool = False
    ) -> None:
        """
        Used to get the paths to the images and segmentations folders. Also used to check if either the series
        descriptions or the path to the series description json dictionary is None.

        Parameters
        ----------
        path_to_patients_folder : str
            The path to the folder that contains all the patients' folders.
        series_descriptions : Optional[Union[str, Dict[str, List[str]]]], default = None.
            A dictionary that contains the series descriptions of the images that needs to be extracted from the
            patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. Note that it can be specified as a path to a json dictionary that contains the
            series descriptions.
        transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
            A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on
            the SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that
            represents the image. DataTransform transforms the data using other a patient's other images or
            segmentations. The keys for images are assumed to be the arbitrary series key set in 'series_descriptions'.
            For segmentation, keys are organ names. Note that if 'series_descriptions' is None, the keys for images are
            assumed to be modalities.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with EXTREME caution!
        """
        self._path_to_patients_folder = path_to_patients_folder
        self._erase_unused_dicom_files = erase_unused_dicom_files
        self._transforms = self._validate_transforms(transforms)

        if isinstance(series_descriptions, str):
            self.path_to_series_description_json = series_descriptions
        elif isinstance(series_descriptions, dict):
            self.series_descriptions = series_descriptions
            self._path_to_series_description_json = None
        elif series_descriptions is None:
            self._series_descriptions = None
            self._path_to_series_description_json = None
        else:
            raise TypeError(
                f"Given series descriptions {series_descriptions} doesn't have the right type. Allowed types are str, "
                f"dict and None."
            )

        self._current_index = 0
        self._patients_who_failed = []

    def __len__(self) -> int:
        """
        Total number of patients.

        Returns
        -------
        length: int
            Total number of patients.
        """
        return len(self.paths_to_patients_folders)

    @property
    def paths_to_patients_folders(self) -> List[str]:
        """
        Get a list of paths to the patients' folders.

        Returns
        -------
        paths_to_patients_folders : List[str]
            A list of paths to the folders containing the patients' data.
        """
        paths_to_folders = []
        for patient_folder_name in os.listdir(self._path_to_patients_folder):
            path_to_folder = os.path.join(
                self._path_to_patients_folder,
                patient_folder_name
            )
            paths_to_folders.append(path_to_folder)

        return paths_to_folders

    @property
    def series_descriptions(self) -> Dict[str, List[str]]:
        """
        Series descriptions.

        Returns
        -------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        """
        return self._series_descriptions

    @series_descriptions.setter
    def series_descriptions(self, series_descriptions: Dict[str, List[str]]) -> None:
        """
        Series descriptions setter.

        Parameters
        ----------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        """
        items = list(series_descriptions.items())
        for previous_items, current_items in zip(items, items[1:]):
            set_intersection = set(previous_items[1]) & set(current_items[1])

            if bool(set_intersection):
                raise AssertionError(
                    f"\nThe dictionary of series descriptions should not contain the same series names for different "
                    f"images/modalities. \nHowever, here we find the series names {previous_items[1]} for the "
                    f"{previous_items[0]} image and {current_items[1]} for the {current_items[0]} image. \nClearly, "
                    f"the images series values are overlapping because of the series named {set_intersection}."
                )

        self._series_descriptions = series_descriptions

    @property
    def path_to_series_description_json(self) -> Union[str, None]:
        """
        Path to series description json.

        Returns
        -------
        path_to_series_description_json : Union[str, None]
            Path to the json dictionary that contains the series descriptions.
        """
        return self._path_to_series_description_json

    @path_to_series_description_json.setter
    def path_to_series_description_json(self, path_to_series_description_json: str) -> None:
        """
        Path to series description json setter. This is used to set the series descriptions attribute using the
        dictionary read from the json file.

        Parameters
        ----------
        path_to_series_description_json : str
            Path to the json dictionary that contains the series descriptions.
        """
        with open(path_to_series_description_json, "r") as json_file:
            self.series_descriptions = deepcopy(json.load(json_file))

        self._path_to_series_description_json = path_to_series_description_json

    @property
    def patients_who_failed(self) -> List[PatientWhoFailed]:
        """
        List of patients with one or more images not added to the HDF5 dataset due to the absence of the series in
        the patient record.

        Returns
        -------
        patients_who_failed : List[PatientWhoFailed]
            List of patients with one or more images not added to the HDF5 dataset due to the absence of the series in
            the patient record.
        """
        return self._patients_who_failed

    @staticmethod
    def _validate_transforms(
            transforms: Union[Compose, DataTransform, PhysicalSpaceTransform, MonaiMapTransform]
    ) -> Optional[Union[Compose, DataTransform, PhysicalSpaceTransform, MonaiMapTransform]]:
        """
        Validate transforms type and set allow_missing_keys attributes to True.

        Parameters
        ----------
        transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
            A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on
            the SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that
            represents the image. DataTransform transforms the data using other a patient's other images or
            segmentations. The keys for images are assumed to be the arbitrary series key set in 'series_descriptions'.
            For segmentation, keys are organ names. Note that if 'series_descriptions' is None, the keys for images are
            assumed to be modalities.

        Returns
        -------
        transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
            A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on
            the SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that
            represents the image. DataTransform transforms the data using other a patient's other images or
            segmentations. The keys for images are assumed to be the arbitrary series key set in 'series_descriptions'.
            For segmentation, keys are organ names. Note that if 'series_descriptions' is None, the keys for images are
            assumed to be modalities.
        """
        if transforms is None:
            return transforms
        elif isinstance(transforms, Compose):
            for t in transforms.transforms:
                if not isinstance(t, (DataTransform, PhysicalSpaceTransform, MonaiMapTransform)):
                    raise AssertionError(
                        "The given transforms must inherit from 'DataTransform', 'PhysicalSpaceTransform' or "
                        "'MonaiMapTransform'."
                    )
                t.allow_missing_keys = True
            return transforms
        elif isinstance(transforms, (DataTransform, PhysicalSpaceTransform, MonaiMapTransform)):
            transforms.allow_missing_keys = True
            return transforms
        else:
            raise AssertionError(
                "'transforms' must either be of type 'Compose', 'DataTransform', 'PhysicalSpaceTransform' or "
                "'MonaiMapTransform'."
            )

    def save_series_descriptions_to_json(self, path: str) -> None:
        """
        Saves the dictionary of series descriptions in a json format at the given path.

        Parameters
        ----------
        path : str
            Path to the json dictionary that will contain the series descriptions.
        """
        with open(path, 'w', encoding='utf-8') as json_file:
            json.dump(self.series_descriptions, json_file, ensure_ascii=False, indent=4)

    def reset(self) -> None:
        """
        Resets the generator.
        """
        self._current_index = 0
        self._patients_who_failed = []

    def send(self, _) -> PatientDataModel:
        """
        Resumes the execution and sends a value into the generator function. This method returns the next value yielded
        by the generator and update the current index or raises StopIteration (via the self.throw method) if all patient
        datasets have been generated.

        Returns
        -------
        patient_dataset : PatientDataModel
            A named tuple grouping the patient's data extracted from its dicom files and the patient's medical image
            segmentation data extracted from the segmentation files.
        """
        if self._current_index == len(self):
            self.throw()

        _logger.info(f"Downloading Patient {self._current_index + 1}/{len(self)}")
        _logger.info(f"Path to patient folder : {self.paths_to_patients_folders[self._current_index]}")

        patient_data_reader = PatientDataReader(
            path_to_patient_folder=self.paths_to_patients_folders[self._current_index],
            series_descriptions=self.series_descriptions,
            erase_unused_dicom_files=self._erase_unused_dicom_files
        )

        if self._series_descriptions is not None:
            self.series_descriptions = patient_data_reader.series_descriptions
        if self.path_to_series_description_json:
            self.save_series_descriptions_to_json(path=self._path_to_series_description_json)
        if patient_data_reader.failed_images:
            failed_images = {image: self.series_descriptions[image] for image in patient_data_reader.failed_images}

            self._patients_who_failed.append(
                PatientWhoFailed(
                    id=patient_data_reader.patient_id,
                    failed_images=failed_images,
                    available_series_descriptions=patient_data_reader.available_series_descriptions
                )
            )

        self._current_index += 1

        return patient_data_reader.get_patient_dataset(transforms=self._transforms)

    def throw(self, typ: Exception = StopIteration, value=None, traceback=None) -> None:
        """
        Raises an exception of type typ.
        """
        raise typ



import logging

from .databases import PatientsDatabase
from .extractors import PatientsDataExtractor, PatientWhoFailed
from .utils import PatientDataModel

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)
logging.getLogger(__name__).addHandler(stream_handler)

__author__ = "Maxence Larose"
__version__ = "1.1.1"
__copyright__ = "Copyright 2022, Maxence Larose"
__credits__ = ["Maxence Larose"]
__license__ = "Apache License 2.0"
__maintainer__ = "Maxence Larose"
__email__ = "maxence.larose.1@ulaval.ca"
__status__ = "Production"



"""
    @file:              dicom_segmentation_factory.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 03/2022

    @Description:       This file contains the abstract class BaseDicomSegmentationFactory and all factories that
                        inherit from this class.
"""

from abc import abstractmethod
from typing import List

import numpy as np
import pydicom
import pydicom_seg
from rt_utils import RTStruct
import SimpleITK as sitk

from .base_segmentation_factory import BaseSegmentationFactory
from ....utils.data_model import ImageDataModel
from .segment import Segment


class BaseDicomSegmentationFactory(BaseSegmentationFactory):
    """
    An abstract class used as a reference for all other segmentation classes that read data from DICOM files. These
    classes differ by the type of segmentation they are able to read. Indeed, the format of the segmentations may be
    different even if the file extension is the same, because there is multiple ways of representing a segment.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        super().__init__(
            image=image,
            path_to_segmentation=path_to_segmentation
        )

        self._dicom: pydicom.FileDataset = pydicom.dcmread(path_to_segmentation)

    @abstractmethod
    def segments(self) -> List[Segment]:
        """
        Segments property.

        Returns
        -------
        segments : List[Segment]
            List of all the segments.
        """
        raise NotImplementedError


class DicomSEGSegmentationFactory(BaseDicomSegmentationFactory):
    """
    Class that defined the methods that are used to get the segments for the DICOM-SEG type of segmentation.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        super().__init__(
            image=image,
            path_to_segmentation=path_to_segmentation
        )

    @property
    def _reader(self) -> pydicom_seg.SegmentReader:
        """
        Reader property.

        Returns
        -------
        reader : pydicom_seg.SegmentReader
            DICOM-SEG reader.
        """
        return pydicom_seg.SegmentReader()

    @property
    def segments(self) -> List[Segment]:
        """
        Segments property.

        Returns
        -------
        segments : List[Segment]
            List of all the segments.
        """
        result = self._reader.read(self._dicom)

        segments = []
        for segment_number, dicom_header in result.segment_infos.items():
            if dicom_header.__contains__("SegmentLabel"):
                organ_name = dicom_header.SegmentLabel
            else:
                organ_name = dicom_header.SegmentDescription

            simple_itk_label_map = result.segment_image(segment_number)

            segments.append(Segment(name=organ_name, simple_itk_label_map=simple_itk_label_map))

        return segments


class RTStructSegmentationFactory(BaseDicomSegmentationFactory):
    """
    Class that defined the methods that are used to get the segments for the RT Struct type of segmentation.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        super().__init__(
            image=image,
            path_to_segmentation=path_to_segmentation
        )

    @property
    def _reader(self) -> RTStruct:
        """
        Reader property.

        Returns
        -------
        reader : RTStruct
            RT Struct reader.
        """
        return RTStruct(
            series_data=[pydicom.dcmread(path) for path in self._image.paths_to_dicoms],
            ds=pydicom.dcmread(self._path_to_segmentation)
        )

    @property
    def segments(self) -> List[Segment]:
        """
        Segments property.

        Returns
        -------
        segments : List[Segment]
            List of all the segments.
        """
        segment_names = self._reader.get_roi_names()
        segments = []

        for organ_name in segment_names:
            array = self._reader.get_roi_mask_by_name(organ_name)
            array = np.multiply(array, 1)
            array = array.transpose(2, 0, 1)

            simple_itk_label_map = sitk.GetImageFromArray(array)
            simple_itk_label_map.SetOrigin(self._image.simple_itk_image.GetOrigin())
            simple_itk_label_map.SetSpacing(self._image.simple_itk_image.GetSpacing())
            simple_itk_label_map.SetDirection(self._image.simple_itk_image.GetDirection())

            segments.append(Segment(name=organ_name, simple_itk_label_map=simple_itk_label_map))

        return segments



"""
    @file:              base_patient_data_factory.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 03/2022

    @Description:       This file contains the class BasePatientDataFactory that is used as an abstract class used as a
                        reference for all other patient data factories.
"""

from abc import ABC, abstractmethod
from os import remove
from typing import Dict, List, Optional

import pydicom

from delia.readers.image.dicom_reader import DicomReader
from delia.utils.data_model import ImageDataModel, PatientDataModel


class BasePatientDataFactory(ABC):
    """
    An abstract class that is used as a reference for all other patient data factories.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            paths_to_segmentations: Optional[List[str]],
            series_descriptions: Optional[Dict[str, List[str]]],
            erase_unused_dicom_files: bool = False
    ):
        """
        Constructor of the class BasePatientDataFactory.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's image files.
        paths_to_segmentations : Optional[List[str]]
            List of paths to the patient's segmentation files.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        self._path_to_patient_folder = path_to_patient_folder
        self._paths_to_segmentations = paths_to_segmentations
        self._series_descriptions = series_descriptions
        self._erase_unused_dicom_files = erase_unused_dicom_files

        dicom_reader = DicomReader(path_to_patient_folder=self._path_to_patient_folder)
        self._images_data = dicom_reader.get_images_data(remove_segmentations=True)

    @property
    def patient_id(self) -> str:
        """
        Patient ID.

        Returns
        -------
        patient_id : str
            Patient ID.
        """
        patient_id = self._images_data[0].dicom_header.PatientID

        return str(patient_id)

    @staticmethod
    def get_segmentation_reference_uid(segmentation_header: pydicom.dataset.FileDataset) -> pydicom.DataElement:
        """
        Get the reference UID of the medical image on which the given segmentation/contouring was performed.

        Parameters
        ----------
        segmentation_header : pydicom.dataset.FileDataset
            Loaded DICOM dataset.

        Returns
        -------
        reference_uid : pydicom.DataElement
            Reference UID of the medical image on which the contouring was performed.
        """
        if hasattr(segmentation_header, "ReferencedSeriesSequence"):
            return segmentation_header.ReferencedSeriesSequence[0].SeriesInstanceUID
        elif hasattr(segmentation_header, "ReferencedFrameOfReferenceSequence"):
            referenced_frame_of_reference_sequence = segmentation_header.ReferencedFrameOfReferenceSequence
            rt_referenced_study_sequence = referenced_frame_of_reference_sequence[0].RTReferencedStudySequence
            rt_referenced_series_sequence = rt_referenced_study_sequence[0].RTReferencedSeriesSequence
            return rt_referenced_series_sequence[0].SeriesInstanceUID
        else:
            raise AssertionError(
                "The segmentation DICOM header must contain either the 'ReferencedSeriesSequence' attribute or "
                "the 'ReferencedFrameOfReferenceSequence' attribute to associate the segmentation with the "
                "corresponding medical image."
            )

    @staticmethod
    def erase_dicom_files(image: ImageDataModel):
        """
        Erases the dicom files associated to a given image.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        """
        for path in image.paths_to_dicoms:
            remove(path)

    @abstractmethod
    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        raise NotImplementedError



"""
    @file:              matching_centroid_spatial_crop.py
    @Author:            Maxence Larose

    @Creation Date:     03/2023
    @Last modification: 03/2023

    @Description:       This file contains the MatchingCentroidSpatialCropd transform.
"""

from typing import Collection, Dict, Hashable, Sequence, Union
from warnings import warn

from monai.transforms import SpatialCrop
import numpy as np

from .tools import compute_centroid
from ..tools import Mode
from .transform import ArraySpaceTransform

KeysCollection = Union[Collection[Hashable], Hashable]


class MatchingCentroidSpatialCropd(ArraySpaceTransform):
    """
    Performs crop around the centroid of a segmentation, get the used coordinates of the spatial bounding box and apply
    this crop on other matching images.

    WARNING! The first segmentation (that has the required key) contained in the list of segmentations of a given image
    is used to compute the centroid. For example, if there is multiple 'Prostate' segmentation of the same CT scan, the
    first segmentation in the files is used to compute the centroid.
    """

    def __init__(
            self,
            segmentation_key: str,
            matching_keys: KeysCollection,
            roi_size: Union[Sequence[int], int]
    ):
        """
        Initializes output spacing.

        Parameters
        ----------
        segmentation_key : str
            Key of the segmentation from which to crop around the centroid and to get coordinates of the spatial
            bounding box. Segmentation keys are organ names.
        matching_keys : KeysCollection
            Keys of the corresponding items to be transformed using the calculated coordinates of the spatial bounding
            box around the centroid of the segmentation. Image keys are assumed to be arbitrary series keys defined in
            'series_descriptions'. For the label maps, the keys are organ names. Note that if 'series_descriptions' is
            None, the image keys are assumed to be modality names.
        roi_size : Union[Sequence[int], int]
            The size of the crop region e.g. [224,224,128]. If a dimension of ROI size is larger than image size, will
            not crop that dimension of the image. If its components have non-positive values, the corresponding size of
            input image will be used. For example: if the spatial size of input data is [40, 40, 40] and
            `roi_size=[32, 64, -1]`, the spatial size of output data will be [32, 40, 40].
        """
        keys = [key for key in matching_keys] + [segmentation_key]
        super().__init__(keys=keys)

        self._centroid = None
        self._segmentation_key = segmentation_key
        self._roi_size = roi_size

    def __call__(self, data: Dict[Hashable, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Performs crop around the centroid of a segmentation, get the used coordinates of the spatial bounding box and
        apply this crop on other matching images.

        Parameters
        ----------
        data : Dict[Hashable, np.ndarray]
            A Python dictionary that contains numpy ndarray.

        Returns
        -------
        transformed_data : Dict[Hashable, np.ndarray]
            A Python dictionary that contains transformed numpy ndarray.
        """
        d = dict(data)

        if self._mode == Mode.NONE:
            raise AssertionError("Transform mode must be set before __call__.")
        elif self._mode == Mode.IMAGE:
            assert self._centroid is not None, "'centroid' must be set before __call__ with image mode."
        elif self._mode == Mode.SEGMENTATION:
            if self._centroid is None:
                self._centroid = compute_centroid(d[self._segmentation_key])
        else:
            raise ValueError("Unknown transform mode.")

        spatial_crop = SpatialCrop(roi_center=self._centroid, roi_size=self._roi_size)

        for key in self.key_iterator(d):
            if key == self._segmentation_key:
                n_original_voxels = (d[key] == 1).sum().item()
                d[key] = spatial_crop(d[key])
                n_final_voxels = (d[key] == 1).sum().item()

                if n_original_voxels != n_final_voxels:
                    warn(
                        f"The matching centroid spatial crop bounding box doesn't contain all the '1' voxels of the "
                        f"orignal label map. The total number of voxels in the original label map was "
                        f"{n_original_voxels} and the total number of voxels in the cropped label map is "
                        f"{n_final_voxels}. The difference (in voxels) is {n_original_voxels - n_final_voxels}."
                    )
            else:
                d[key] = spatial_crop(d[key])

        if self._mode == Mode.IMAGE:
            self._centroid = None

        return d


MatchingCentroidSpatialCropD = MatchingCentroidSpatialCropDict = MatchingCentroidSpatialCropd



from datetime import datetime
import logging
import logging.config
import os
import sys
import yaml


# Append module root directory to sys.path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def configure_logging(path_to_configuration_file: str):
    now = datetime.now()
    logs_dir = f"logs/{now.strftime('%Y-%m-%d')}"
    logs_file = f"{logs_dir}/{now.strftime('%Y-%m-%d_%H-%M-%S')}.log"
    os.makedirs(logs_dir, exist_ok=True)

    with open(path_to_configuration_file, 'r') as stream:
        config: dict = yaml.load(stream, Loader=yaml.FullLoader)

    config["handlers"]["file"]["filename"] = logs_file

    logging.config.dictConfig(config)



"""
    @file:              utils.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 03/2022

    @Description:       A collection of functions and classes that may or may not be useful.
"""

import os
from typing import Callable
import warnings


def is_path_valid(
        path: str
) -> None:
    """
    Raise a ValueError if the given path doesn't exist.

    Parameters
    ----------
    path : str
        A path.

    Returns
    -------
    None
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"Given path {path} does not exist.")


class Decorators:

    @classmethod
    def deprecated(cls, explanation: str) -> Callable:
        """
        Decorator to use on deprecated functions.

        Parameters
        ----------
        explanation : str
            Explanation of depreciation and recommended alternative function.

        Returns
        -------
        deprecated_func: Callable
            Deprecated function.
        """
        def deprecated_func(func):
            def wrapper(*args, **kwargs):
                warnings.warn(explanation, DeprecationWarning)
                return func(*args, **kwargs)

            return wrapper

        return deprecated_func



"""
    @file:              segmentation_factory.py
    @Author:            Maxence Larose

    @Creation Date:     12/2021
    @Last modification: 01/2022

    @Description:       This file contains the class SegmentationFactory that is used as an abstract class used as a
                        reference for all other segmentation classes that read data and build a Segmentation object
                        from it.
"""

from abc import ABC, abstractmethod
from typing import List

from ....utils.data_model import ImageDataModel
from .segment import Segment
from .segmentation import Segmentation


class BaseSegmentationFactory(ABC):
    """
    An abstract class that is used as a reference for all other segmentation classes that read data and build a
    Segmentation object from it.
    """

    def __init__(
            self,
            image: ImageDataModel,
            path_to_segmentation: str,
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        image : ImageDataModel
            A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of
            the paths to each dicom contained in the series.
        path_to_segmentation : str
            The path to the segmentation file.
        """
        self._image = image
        self._path_to_segmentation = path_to_segmentation

    @property
    @abstractmethod
    def segments(self) -> List[Segment]:
        """
        Abstract segments property.

        Returns
        -------
        segments : List[Segment]
            List of all the segments.
        """
        pass

    def create_segmentation(self) -> Segmentation:
        """
        Creates a Segmentation using the list of Segments.

        Returns
        -------
        segmentation : Segmentation
            Segmentation.
        """
        return Segmentation(segments=self.segments)



"""
    @file:              transform.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 11/2022

    @Description:       This file contains the PhysicalSpaceTransform abstract class which is used to define transforms
                        that can be applied to images and segmentations.
"""

from abc import abstractmethod
from typing import Collection, Dict, Hashable, NamedTuple, Union

import SimpleITK as sitk
import pydicom
from monai.transforms import MapTransform

from ..tools import Mode

KeysCollection = Union[Collection[Hashable], Hashable]


class ImageData(NamedTuple):
    """
    A named tuple the medical image as a simpleITK image and its dicom header.

    Elements
    --------
    simple_itk_image : Image
        Segmentation as a SimpleITK image.
    dicom_header : FileDataset
        Dicom header dataset.
    """
    simple_itk_image: sitk.Image
    dicom_header: pydicom.dataset.FileDataset = None


class PhysicalSpaceTransform(MapTransform):
    """
    PhysicalSpaceTransform abstract class.
    """

    def __init__(self, keys: KeysCollection) -> None:
        """
        Initialize transform keys.

        Parameters
        ----------
        keys : KeysCollection
            Keys of the corresponding items to be transformed. Image keys are assumed to be arbitrary series keys
            defined in 'series_descriptions'. For the label maps, the keys are organ names. Note that if
            'series_descriptions' is None, the image keys are assumed to be modality names.
        """
        super().__init__(keys=keys, allow_missing_keys=True)
        self._mode = Mode.NONE

    @property
    def mode(self) -> Mode:
        """
        The transform mode, i.e. whether the single ITK images on which to apply the transformation are images OR
        segmentations.

        Returns
        -------
        mode : Mode
            Transform mode.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: Mode):
        """
        The transform mode, i.e. whether the single ITK images on which to apply the transformation are images OR
        segmentations.

        Parameters
        ----------
        mode : Mode
            The transform mode, i.e. whether the single ITK images on which to apply the transformation are images OR
            segmentations.
        """
        self._mode = mode

    @abstractmethod
    def __call__(self, data: Dict[str, ImageData]) -> Dict[Hashable, sitk.Image]:
        """
        Apply the transformation.

        Parameters
        ----------
        data : Dict[str, ImageData]
            A Python dictionary that contains ImageData.

        Returns
        -------
        transformed_data : Dict[Hashable, sitk.Image]
            A Python dictionary that contains transformed SimpleITK images.
        """
        raise NotImplementedError



"""
    @file:              segmentation.py
    @Author:            Maxence Larose

    @Creation Date:     12/2021
    @Last modification: 12/2021

    @Description:       This file contains the class Segmentation.
"""

from typing import Dict, List

import SimpleITK as sitk

from .segment import Segment


class Segmentation:
    """
    A class used as a reference for all other abstract segmentation classes. These classes will differ by the extension
    of the segmentation files they are able to read.
    """

    def __init__(self, segments: List[Segment]):
        """
        Constructor of the segmentation object.

        Parameters
        ----------
        segments : List[Segment]
            List of all the segments.
        """
        self._segments = segments

    def __len__(self) -> int:
        """
        Number of segments in the segmentation.

        Returns
        -------
        number_of_segments : int
            Number of segments
        """
        return len(self._segments)

    def __getitem__(self, index: int) -> Segment:
        """
        Get a specific segment using its index.

        Parameters
        ----------
        index : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        return self._segments[index]

    def __setitem__(self, index: int, value) -> None:
        """
        Set a segment using its index and its value.

        Parameters
        ----------
        index : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        self._segments[index] = value

    @property
    def simple_itk_label_maps(self) -> Dict[str, sitk.Image]:
        """
        SimpleITK label maps of all the organs/segments that are found in the segmentation file.

        Returns
        -------
        simple_itk_label_maps : Dict[str, sitk.Image]
            A dictionary that contains the name of the organs and their corresponding binary label map as a simpleITK
            image. Keys are organ names and values are binary label maps. Thus, the label maps dictionary is formatted
            as follows :

                label_maps = {
                    organ_name (example: "PROSTATE"): sitk.Image,
                    organ_name (example: "RECTUM"): sitk.Image,
                    ...
                }
        """
        return {segment.name: segment.simple_itk_label_map for segment in self._segments}



"""
    @file:              patients_database.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains the PatientDatabase class that is used to interact with an hdf5 file
                        database. The main purpose of this class is to create an hdf5 file database from multiple
                        patients dicom files and their segmentation. This class also allows the user to interact with
                        an existing hdf5 file database through queries.
"""

import logging
import os
from typing import List, Optional, Tuple, Union

import h5py
import json
import numpy as np
import SimpleITK as sitk

from delia.extractors.patients_data_extractor import PatientsDataExtractor, PatientWhoFailed
from delia.utils.data_model import ImageAndSegmentationDataModel

_logger = logging.getLogger(__name__)


class PatientsDatabase:
    """
    A class that is used to interact with a patients database. The main purpose of this class is to create an hdf5 file
    database from multiple patients dicom files and their segmentation. This class also allows the user to interact with
    an existing hdf5 file database through queries.
    """

    DICOM_HEADER = "Dicom_header"
    IMAGE = "Image"
    MODALITY = "Modality"
    TRANSFORMS = "Transforms"

    def __init__(
            self,
            path_to_database: str,
    ):
        """
        Used to initialize the path to the database.

        Parameters
        ----------
        path_to_database : str
            Path to database.
        """
        self.path_to_database = path_to_database

        if os.path.exists(path_to_database):
            self._file = h5py.File(path_to_database, mode="r")
        else:
            self._file = None

    @property
    def path_to_database(self) -> str:
        """
        Path to database.

        Returns
        -------
        path_to_database : str
            Path to database containing modality and organ names.
        """
        return self._path_to_database

    @path_to_database.setter
    def path_to_database(self, path_to_database: str) -> None:
        """
        Path to database.

        Parameters
        ----------
        path_to_database : str
            Path to database.
        """
        if path_to_database.endswith(".h5"):
            self._path_to_database = path_to_database
        else:
            self._path_to_database = f"{path_to_database}.h5"

    def __getitem__(self, patient: Union[int, List[int], str, List[str]]) -> Union[List[h5py.Group], h5py.Group]:
        """
        Gets a patient group given the patient ID. This method returns a copy of the patient group.

        Parameters
        ----------
        patient : Union[int, List[int], str, List[str]]
            Patient ID (strings) or index (integers).
        """
        if os.path.exists(self.path_to_database):
            if isinstance(patient, int):
                file_keys = list(self._file.keys())
                patient = file_keys[patient]
            elif isinstance(patient, list):
                file_keys = list(self._file.keys())
                patient = [file_keys[p] if isinstance(p, int) else p for p in patient]

            if isinstance(patient, str):
                return self._file[patient]
            elif isinstance(patient, list):
                return [self._file[uid] for uid in patient]
            else:
                raise AssertionError(
                    f"Patient ID should be a list of patient ids (List[str]) or a single patient id (str). "
                    f"Received {type(patient)}."
                )
        else:
            raise AssertionError(f"Database with path {self.path_to_database} doesn't exist. Use 'create'.")

    def __len__(self) -> int:
        """
        Gets number of patients in the database of h5py.File.

        Returns
        -------
        length : int
            Number of patients in the database.
        """
        if self._file:
            return len(self._file.keys())
        else:
            return 0

    def _check_authorization_of_database_creation(
            self,
            overwrite_database: bool
    ) -> None:
        """
        Check if database's creation is allowed.

        Parameters
        ----------
        overwrite_database : bool
            Overwrite existing database.
        """
        if os.path.exists(self.path_to_database):
            if not overwrite_database:
                raise FileExistsError(
                    "The database already exists. You may overwrite it using overwrite_database = True."
                )
            else:
                self.close()
                _logger.info(f"Overwriting HDF5 database with path : {self.path_to_database}")
        else:
            _logger.info(f"Writing HDF5 database with path : {self.path_to_database}")

    @staticmethod
    def _add_sitk_image_attributes_to_hdf5_group(
            patient_image_data: ImageAndSegmentationDataModel,
            group: h5py.Group
    ) -> None:
        """
        Add Simple ITK image information as attributes in the given HDF5 group.

        Parameters
        ----------
        patient_image_data : ImageAndSegmentationDataModel
            A named tuple grouping the patient data retrieved from his dicom files and the segmentation data retrieved
            from the segmentation file.
        group : h5py.Group
            An hdf5 group.
        """
        group.attrs.create(name="Size", data=patient_image_data.image.simple_itk_image.GetSize())
        group.attrs.create(name="Origin", data=patient_image_data.image.simple_itk_image.GetOrigin())
        group.attrs.create(name="Spacing", data=patient_image_data.image.simple_itk_image.GetSpacing())
        group.attrs.create(name="Direction", data=patient_image_data.image.simple_itk_image.GetDirection())
        group.attrs.create(name="Pixel Type", data=patient_image_data.image.simple_itk_image.GetPixelIDTypeAsString())

    @staticmethod
    def _add_dicom_attributes_to_hdf5_group(
            patient_image_data: ImageAndSegmentationDataModel,
            group: h5py.Group,
            tags_to_use_as_attributes: List[Tuple[int, int]],
    ) -> None:
        """
        Add the specified DICOM tags as attributes in the given HDF5 group.

        Parameters
        ----------
        patient_image_data : ImageAndSegmentationDataModel
            A named tuple grouping the patient data retrieved from his dicom files and the segmentation data retrieved
            from the segmentation file.
        group : h5py.Group
            An hdf5 group.
        tags_to_use_as_attributes : List[Tuple[int, int]]
            List of DICOM tags to add as series attributes in the HDF5 database.
        """
        for tag in tags_to_use_as_attributes:
            dicom_data_element = patient_image_data.image.dicom_header[tag]

            if type(dicom_data_element.value) == str:
                data = dicom_data_element.value
            else:
                data = dicom_data_element.repval

            group.attrs.create(name=dicom_data_element.name, data=data)

    @staticmethod
    def _transpose(array: np.ndarray) -> np.ndarray:
        """
        Transpose an array if its shape is not valid for the hdf5 database format.

        Parameters
        ----------
        array : np.ndarray
            An numpy array.

        Returns
        -------
        transposed_array : np.ndarray
            The original array or the transposed array depending on its input shape.
        """
        return array.transpose((1, 2, 0))

    def create(
            self,
            patients_data_extractor: PatientsDataExtractor,
            tags_to_use_as_attributes: Optional[List[Tuple[int, int]]] = None,
            add_sitk_image_metadata_as_attributes: bool = True,
            organs_to_keep: Optional[Union[str, List[str]]] = None,
            overwrite_database: bool = False
    ) -> List[PatientWhoFailed]:
        """
        Create an hdf5 file database from multiple patients dicom files and their segmentation. The goal is to create
        an object from which it is easier to obtain patient images and their segmentation than separated dicom files
        and segmentation files.

        Parameters
        ----------
        patients_data_extractor : PatientsDataExtractor
            An object used to iterate on multiple patients' dicom files and segmentation files using the
            PatientDataReader to obtain all patients' data.
        tags_to_use_as_attributes : List[Tuple[int, int]]
            List of DICOM tags to add as series attributes in the HDF5 database.
        add_sitk_image_metadata_as_attributes : bool, default = True.
            Keep Simple ITK image information as attributes in the corresponding series.
        organs_to_keep : Optional[Union[str, List[str]]]
            Organ segmentations to keep in the database. By default, all organs are kept.
        overwrite_database : bool, default = False.
            Overwrite existing database.

        Returns
        -------
        patients_who_failed : List[PatientWhoFailed]
            List of patients with one or more images not added to the HDF5 database due to the absence of the series in
            the patient record.
        """
        self._check_authorization_of_database_creation(overwrite_database=overwrite_database)

        if tags_to_use_as_attributes is None:
            tags_to_use_as_attributes = []
        if isinstance(organs_to_keep, str):
            organs_to_keep = [organs_to_keep]

        with h5py.File(self.path_to_database, "w") as file:
            number_of_patients = len(patients_data_extractor)
            for patient_idx, patient_dataset in enumerate(patients_data_extractor):
                patient_id = patient_dataset.patient_id
                patient_group = file.create_group(name=patient_id)

                for image_idx, patient_image_data in enumerate(patient_dataset.data):
                    series_group = patient_group.create_group(name=str(image_idx))

                    self._add_dicom_attributes_to_hdf5_group(
                        patient_image_data, series_group, tags_to_use_as_attributes
                    )

                    if add_sitk_image_metadata_as_attributes:
                        self._add_sitk_image_attributes_to_hdf5_group(patient_image_data, series_group)

                    series_group.create_dataset(
                        name=self.DICOM_HEADER,
                        data=json.dumps(patient_image_data.image.dicom_header.to_json_dict())
                    )

                    image_array = sitk.GetArrayFromImage(patient_image_data.image.simple_itk_image)

                    series_group.create_dataset(
                        name=self.IMAGE,
                        data=self._transpose(image_array)
                    )

                    if patient_image_data.segmentations:
                        for segmentation_idx, segmentation in enumerate(patient_image_data.segmentations):
                            segmentation_group = series_group.create_group(name=str(segmentation_idx))
                            segmentation_group.attrs.create(name=self.MODALITY, data=segmentation.modality)

                            for organ, simple_itk_label_map in segmentation.simple_itk_label_maps.items():
                                numpy_array_label_map = sitk.GetArrayFromImage(simple_itk_label_map)

                                if organs_to_keep is None or organ in organs_to_keep:
                                    segmentation_group.create_dataset(
                                        name=organ,
                                        data=self._transpose(numpy_array_label_map),
                                        dtype=np.int8
                                    )

                for idx, transform in enumerate(patient_dataset.transforms_history.history):
                    patient_group.attrs.create(
                        name=f"{self.TRANSFORMS}_{idx}",
                        data=json.dumps(
                            obj=transform,
                            default=patient_dataset.transforms_history.serialize
                        )
                    )

                _logger.info(f"Progress : {patient_idx + 1}/{number_of_patients} patients added to database.")

            patients_data_extractor.close()
            patients_data_extractor.reset()

        self._file = h5py.File(self.path_to_database, "r")

        return patients_data_extractor.patients_who_failed

    def close(self):
        """
        Closes the hdf5 file database.
        """
        if self._file:
            self._file.close()

    def __del__(self):
        """
        Closes the hdf5 file database.
        """
        self.close()



"""
    @file:              transform.py
    @Author:            Maxence Larose

    @Creation Date:     11/2022
    @Last modification: 11/2022

    @Description:       This file contains the DataTransform abstract class which is used to define transforms
                        that can be applied to images and segmentations.
"""

from abc import abstractmethod
from typing import Collection, Dict, Hashable, Union

from monai.transforms import MapTransform

from delia.utils.data_model import ImageAndSegmentationDataModel

KeysCollection = Union[Collection[Hashable], Hashable]


class DataTransform(MapTransform):
    """
    DataTransform abstract class.
    """

    def __init__(self, keys: KeysCollection) -> None:
        """
        Initialize transform keys.

        Parameters
        ----------
        keys : KeysCollection
            Keys of the corresponding items to be transformed. Image keys are assumed to be arbitrary series keys
            defined in 'series_descriptions'. For the label maps, the keys are organ names. Note that if
            'series_descriptions' is None, the image keys are assumed to be modality names.
        """
        super().__init__(keys=keys, allow_missing_keys=True)

    @abstractmethod
    def __call__(self, data: Dict[str, ImageAndSegmentationDataModel]) -> Dict[Hashable, ImageAndSegmentationDataModel]:
        """
        Apply the transformation.

        Parameters
        ----------
        data : Dict[str, ImageAndSegmentationDataModel]
            A Python dictionary that contains ImageAndSegmentationDataModel.

        Returns
        -------
        transformed_data : Dict[Hashable, ImageAndSegmentationDataModel]
            A Python dictionary that contains transformed ImageAndSegmentationDataModel.
        """
        raise NotImplementedError



"""
    @file:              data_model.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains several named tuples that are used to standardize the structure of objects
                        containing data.
"""

from dataclasses import dataclass
from typing import Dict, Optional, List, Sequence

import numpy as np
import pydicom
import SimpleITK as sitk

from .transforms_history import TransformsHistory


@dataclass
class SegmentationDataModel:
    """
    A named tuple grouping the segmentation as several binary label maps (one for each organ in the segmentation) and
    the segmentation as a simpleITK image.

    Elements
    --------
    dicom_header : FileDataset
        Dicom header dataset.
    modality : str
        Segmentation modality (ex: SEG, RTSTRUCT).
    simple_itk_label_maps : Dict[str, sitk.Image]
        A dictionary that contains the name of the organs and their corresponding binary label map as a simpleITK
        image. Keys are organ names and values are binary label maps. Thus, the label maps dictionary is formatted
        as follows :

            label_maps = {
                organ_name (example: "PROSTATE"): sitk.Image,
                organ_name (example: "RECTUM"): sitk.Image,
                ...
            }
    """
    dicom_header: pydicom.dataset.FileDataset
    modality: Optional[str] = None
    simple_itk_label_maps: Optional[Dict[str, sitk.Image]] = None

    @property
    def numpy_array_label_maps(self) -> Optional[Dict[str, np.ndarray]]:
        if self.simple_itk_label_maps:
            return {k: sitk.GetArrayFromImage(v) for k, v in self.simple_itk_label_maps.items()}
        else:
            return None


@dataclass
class ImageDataModel:
    """
    A named tuple grouping the patient's dicom header, its medical image as a simpleITK image and a sequence of the
    paths to each dicom contained in the series.

    Elements
    --------
    dicom_header : FileDataset
        Dicom header dataset.
    simple_itk_image : Image
        Dicom data volume as a SimpleITK image.
    paths_to_dicoms : Sequence[str]
        A list of the paths to each dicom contained in the series.
    series_key : str
        Arbitrary name given to the image we want to extract.
    _transforms_key : str
        key used for the temporary dictionary created when applying transforms.
    """
    dicom_header: pydicom.dataset.FileDataset
    paths_to_dicoms: Sequence[str]
    simple_itk_image: sitk.Image
    series_key: Optional[str] = None
    transforms_key: Optional[str] = None

    @property
    def numpy_array(self) -> np.ndarray:
        return sitk.GetArrayFromImage(self.simple_itk_image)


@dataclass
class ImageAndSegmentationDataModel:
    """
    A named tuple grouping the patient data retrieved from his dicom files and the segmentation data retrieved from
    the segmentation file.

    Elements
    --------
    image : ImageDataModel
        The patient's medical image data.
    segmentation : Sequence[SegmentationDataModel]
        Data from the segmentation of the patient's medical image.
    """
    image: ImageDataModel
    segmentations: Optional[Sequence[SegmentationDataModel]] = None


@dataclass
class PatientDataModel:
    """
    A named tuple grouping the patient's data extracted from its dicom files and the patient's medical image
    segmentation data extracted from the segmentation files, for each available modality. The patient data model is
    formatted as follows :

        PatientDataModel = (
            "patient_id": str,
            "data": [
                ImageAndSegmentationDataModel,
                ImageAndSegmentationDataModel
                ...
            ]
            "transforms_history" : TransformsHistory
        )
    """
    patient_id: str
    data: List[ImageAndSegmentationDataModel]
    transforms_history: Optional[TransformsHistory] = None



"""
    @file:              segment.py
    @Author:            Maxence Larose

    @Creation Date:     12/2021
    @Last modification: 12/2021

    @Description:       This file contains the Segment class.
"""

import SimpleITK as sitk


class Segment:
    """
    A class used to represent a segment object.
    """

    def __init__(self, name: str, simple_itk_label_map: sitk.Image):
        """
        Constructor of the segment object.

        Parameters
        ----------
        name : str
            Segment's name.
        simple_itk_label_map : sitk.Image
            The segmentation as a SimpleITK image.
        """
        self._name = name
        self._simple_itk_label_map = simple_itk_label_map

    @property
    def name(self) -> str:
        """
        Name property.

        Returns
        -------
        name : str
            Segment's name.
        """
        return self._name

    @property
    def simple_itk_label_map(self) -> sitk.Image:
        """
        Simple ITK label map image.

        Returns
        -------
        simple_itk_label_map : sitk.Image
            The segmentation as a SimpleITK image.
        """
        return self._simple_itk_label_map



"""
    @file:              dicom_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 03/2022

    @Description:       This file contains the DicomReader class which is used to read dicom files contained in a given
                        folder.
"""

from collections import defaultdict
from glob import glob
import logging
import os
from typing import Dict, List, NamedTuple, Set

import pydicom
import SimpleITK as sitk

from delia.utils.data_model import ImageDataModel
from delia.readers.segmentation.segmentation_strategy import SegmentationStrategies
from delia.utils.tools import is_path_valid

_logger = logging.getLogger(__name__)


class DicomReader:
    """
    A class used to read dicom files contained in a given folder.
    """

    UNKNOWN = "Unknown"

    class SeriesData(NamedTuple):
        """
        Series description namedtuple to simplify management of values.
        """
        series_description: str
        paths_to_dicoms_from_series: List[str]
        dicom_header: pydicom.dataset.FileDataset

    def __init__(
            self,
            path_to_patient_folder: str
    ):
        """
        Constructor of the class DicomReader.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient DICOM files.
        """
        super(DicomReader, self).__init__()

        is_path_valid(path=path_to_patient_folder)
        self._path_to_patient_folder = path_to_patient_folder

        self._images_series_data_dict = {}
        self._segmentations_series_data_dict = {}
        self.__series_data_dict = self.__get_series_data_dict()

    @staticmethod
    def get_dicom_header(
            path_to_dicom: str,
            show: bool = False
    ) -> pydicom.dataset.FileDataset:
        """
        Get a dicom header given the path to the dicom.

        Parameters
        ----------
        path_to_dicom : str
            The path to the dicom file.
        show : bool
            Show dicom header in console.

        Returns
        -------
        loaded_dicom : pydicom.dataset.FileDataset
            Loaded DICOM dataset.
        """
        loaded_dicom = pydicom.dcmread(path_to_dicom, stop_before_pixels=True)

        if show:
            _logger.info(loaded_dicom)

        return loaded_dicom

    @staticmethod
    def _get_series_description(dicom_header: pydicom.FileDataset) -> str:
        """
        Get Series Description of given DICOM header.

        Parameters
        ----------
        dicom_header : pydicom.dataset.FileDataset
            Loaded DICOM dataset.

        Returns
        -------
        series_description : str
            Series Description.
        """
        if hasattr(dicom_header, "SeriesDescription"):
            return dicom_header.SeriesDescription
        else:
            return DicomReader.UNKNOWN

    def __get_paths_to_dicoms_from_series(self, path_to_dicom_folder: str) -> Dict[str, List[str]]:
        """
        Get all series IDs from a patient's dicom folder.

        Parameters
        ----------
        path_to_dicom_folder : str
            Path to a folder containing only DICOM files.

        Returns
        -------
        paths_to_dicoms_from_series : Dict[str, List[str]]
            Dictionary of lists of paths to dicoms from each series.
        """
        paths = [
            y for x in os.walk(path_to_dicom_folder) for y in glob(os.path.join(x[0], '*')) if os.path.isfile(y) is True
        ]

        paths_and_headers_for_each_series = defaultdict(list)
        for path in paths:
            dicom_header = self.get_dicom_header(path_to_dicom=path)
            paths_and_headers_for_each_series[dicom_header.SeriesInstanceUID].append((path, dicom_header))

        paths_to_dicoms_from_series = defaultdict(list)
        for series_id, result in paths_and_headers_for_each_series.items():
            if len(result) != 1:
                if hasattr(result[1][1], "SliceLocation"):
                    paths_to_dicoms_from_series[series_id] = [
                        i[0] for i in sorted(result, key=lambda s: s[1].SliceLocation)
                    ]
                else:
                    paths_to_dicoms_from_series[series_id] = [
                        i[0] for i in result
                    ]
            else:
                paths_to_dicoms_from_series[series_id] = [result[0][0]]

        return paths_to_dicoms_from_series

    def __get_series_data_dict(self) -> Dict[str, SeriesData]:
        """
        Get the series data from series IDs and the path to the patient's dicom folders.

        Returns
        -------
        series_data_dict : Dict[str, SeriesData]
            Dictionary of the data from the selected series.
        """
        series_data_dict: Dict[str, DicomReader.SeriesData] = {}
        all_patient_ids: Set[str] = set()
        for series_id, paths in self.__get_paths_to_dicoms_from_series(self._path_to_patient_folder).items():
            path_to_first_dicom_of_series = paths[0]
            loaded_dicom_header = self.get_dicom_header(path_to_dicom=path_to_first_dicom_of_series)
            all_patient_ids.add(loaded_dicom_header.PatientID)

            series_data = self.SeriesData(
                series_description=self._get_series_description(loaded_dicom_header),
                paths_to_dicoms_from_series=paths,
                dicom_header=loaded_dicom_header
            )
            series_data_dict[series_id] = series_data

        if len(all_patient_ids) == 0:
            raise AssertionError(f"The patient folder must contain DICOM files. This is not the case for the patient "
                                 f"whose data is currently being downloaded since no DICOM files are found in its "
                                 f"folder.")
        elif len(all_patient_ids) > 1:
            raise AssertionError(f"All DICOM files in the same folder must belong to the same patient. This is not "
                                 f"the case for the patient whose data is currently being downloaded since the "
                                 f"identifiers {all_patient_ids} are found in their folder.")

        for series_id, series_data in series_data_dict.items():
            if series_data.dicom_header.Modality in SegmentationStrategies.get_available_modalities():
                self._segmentations_series_data_dict[series_id] = series_data
            else:
                self._images_series_data_dict[series_id] = series_data

        return series_data_dict

    @staticmethod
    def __get_3d_sitk_image_from_dicom_series(
            paths_to_dicoms_from_series: List[str],
    ) -> sitk.Image:
        """
        Get a 3D image array from a list of dicom paths associated to the same series.

        Parameters
        ----------
        paths_to_dicoms_from_series : List[str]
            A list of the paths to each dicom contained in the series.

        Returns
        -------
        image : sitk.Image
            3D SimpleITK image obtained from the series.
        """
        series_reader = sitk.ImageSeriesReader()
        series_reader.SetFileNames(paths_to_dicoms_from_series)

        image = series_reader.Execute()

        return image

    def get_dicom_headers(self, remove_segmentations: bool = False) -> List[pydicom.dataset.FileDataset]:
        """
        List of all the patient's dicom headers.

        Parameters
        ----------
        remove_segmentations : bool
            Whether or not to keep the segmentation files headers, default = False.

        Returns
        -------
        dicom_headers : List[pydicom.dataset.FileDataset]
            A list of the patient's DICOM headers.
        """
        dicom_headers = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):
            if idx == 0:
                _logger.info(f"Patient ID : {series_data.dicom_header.PatientID}")
                _logger.debug(f"Path to images folder : {self._path_to_patient_folder}")

            _logger.debug(f"  Series description : {series_data.series_description}")

            if not remove_segmentations:
                dicom_headers.append(series_data.dicom_header)
            elif remove_segmentations:
                if series_data.dicom_header.Modality not in SegmentationStrategies.get_available_modalities():
                    dicom_headers.append(series_data.dicom_header)

        return dicom_headers

    def get_images_data(
            self,
            remove_segmentations: bool = False
    ) -> List[ImageDataModel]:
        """
        List of tuples containing simpleITK 3D images array and their corresponding dicom header. Each element in the
        list corresponds to an image series.

        Parameters
        ----------
        remove_segmentations : bool
            Whether to keep the segmentation files headers, default = False.

        Returns
        -------
        image_data_list : List[ImageDataModel]
            A list of the patient's images data.
        """
        images_data = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):
            if idx == 0:
                _logger.info(f"Series description found in the patient's images folder :")
            _logger.info(f"  Series description {idx + 1}: {series_data.series_description}")

            if remove_segmentations and (series_data.dicom_header.Modality in
                                         SegmentationStrategies.get_available_modalities()):
                pass
            elif series_data.series_description == DicomReader.UNKNOWN:
                pass
            else:
                try:
                    image = self.__get_3d_sitk_image_from_dicom_series(
                        paths_to_dicoms_from_series=series_data.paths_to_dicoms_from_series
                    )

                    image_data = ImageDataModel(
                        dicom_header=series_data.dicom_header,
                        simple_itk_image=image,
                        paths_to_dicoms=series_data.paths_to_dicoms_from_series
                    )

                    images_data.append(image_data)

                except RuntimeError as e:
                    _logger.error(
                        f"      RuntimeError : {e}. Simple ITK raised an error while loading the series named "
                        f"{series_data.series_description}. This series is therefore ignored."
                    )

        return images_data



"""
    @Title:             On-the-fly tasks.

    @Description:       Perform on-the-fly tasks on images.
"""

import env_examples  # Modifies path, DO NOT REMOVE

from delia.extractors import PatientsDataExtractor
from delia.transforms import Compose, CopySegmentationsD, MatchingCentroidSpatialCropD, PETtoSUVD, ResampleD
from monai.transforms import KeepLargestConnectedComponentD
import SimpleITK as sitk


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #                                         Logs Setup (Optional)                                               #
    # ----------------------------------------------------------------------------------------------------------- #
    env_examples.configure_logging("logging_conf.yaml")

    # ----------------------------------------------------------------------------------------------------------- #
    #                                      Create patients data extractor                                         #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        series_descriptions="data/series_descriptions.json",
        transforms=Compose(
            [
                ResampleD(keys=["CT_THORAX", "Heart"], out_spacing=(1.5, 1.5, 1.5)),
                MatchingCentroidSpatialCropD(
                    segmentation_key="Heart",
                    matching_keys=["CT_THORAX"],
                    roi_size=(96, 96, 96)
                ),
                PETtoSUVD(keys=["PET"]),
                KeepLargestConnectedComponentD(keys=["Heart"]),
                CopySegmentationsD(segmented_image_key="CT_THORAX", unsegmented_image_key="PET")
            ]
        )
    )

    # ----------------------------------------------------------------------------------------------------------- #
    #                                    Perform on-thy-fly tasks on images                                       #
    # ----------------------------------------------------------------------------------------------------------- #
    for patient_data in patients_data_extractor:
        print(f"{'-'*20}\nPatient ID: {patient_data.patient_id}")

        for patient_image_data in patient_data.data:
            dicom_header = patient_image_data.image.dicom_header
            simple_itk_image = patient_image_data.image.simple_itk_image
            numpy_array_image = sitk.GetArrayFromImage(simple_itk_image)

            """Perform any tasks on images on-the-fly."""
            print("\nMODALITY:", dicom_header.Modality)
            print("IMAGE NAME:", patient_image_data.image.series_key)
            print("IMAGE SHAPE:", numpy_array_image.shape)
            print("IMAGE MEAN VALUE:", numpy_array_image.mean())

            segmentations = patient_image_data.segmentations
            if segmentations:
                for segmentation in segmentations:
                    for organ, sitk_label_map in segmentation.simple_itk_label_maps.items():
                        numpy_array_image = sitk.GetArrayFromImage(sitk_label_map)
                        print("ORGAN:", organ)
                        print("LABEL MAP SHAPE:", numpy_array_image.shape)



"""
    @Title:             Radiomics extraction.

    @Description:       Extract radiomics from CT and PET images with a given 'Heart' contour.
"""

import env_examples  # Modifies path, DO NOT REMOVE

from delia.extractors import PatientsDataExtractor
from delia.radiomics import RadiomicsDataset, RadiomicsFeatureExtractor
from delia.transforms import Compose, CopySegmentationsD, PETtoSUVD


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #                                         Logs Setup (Optional)                                               #
    # ----------------------------------------------------------------------------------------------------------- #
    env_examples.configure_logging("logging_conf.yaml")

    # ----------------------------------------------------------------------------------------------------------- #
    #                                      Create patients data extractor                                         #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        series_descriptions="data/radiomics_series_descriptions.json",
        transforms=Compose(
            [
                PETtoSUVD(keys=["PET"]),
                CopySegmentationsD(segmented_image_key="CT_THORAX", unsegmented_image_key="PET")
            ]
        )
    )

    # ----------------------------------------------------------------------------------------------------------- #
    #       Extract radiomics features of the CT image from the segmentation of the heart made on the CT image    #
    # ----------------------------------------------------------------------------------------------------------- #

    CT_radiomics_dataset = RadiomicsDataset(path_to_dataset="data/CT_radiomics")

    CT_radiomics_dataset.extractor = RadiomicsFeatureExtractor(
        path_to_params="features_extractor_params_CT.yaml",
        geometryTolerance=1e-4
    )

    CT_radiomics_dataset.create(
        patients_data_extractor=patients_data_extractor,
        organ="Heart",
        image_name="CT_THORAX"
    )

    # ----------------------------------------------------------------------------------------------------------- #
    #       Extract radiomics features of the PT image from the segmentation of the heart made on the CT image    #
    # ----------------------------------------------------------------------------------------------------------- #
    PT_radiomics_dataset = RadiomicsDataset(path_to_dataset="data/PT_radiomics")

    PT_radiomics_dataset.extractor = RadiomicsFeatureExtractor(
        path_to_params="features_extractor_params_PT.yaml",
        geometryTolerance=1e-4
    )

    PT_radiomics_dataset.create(
        patients_data_extractor=patients_data_extractor,
        organ="Heart",
        image_name="PET"
    )



"""
    @file:              segmentation_strategy.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains the class SegmentationStrategies that enumerates the available categories of
                        segmentation files.
"""

import enum
from typing import List, NamedTuple, Callable

from .factories.dicom_segmentation_factories import (DicomSEGSegmentationFactory, RTStructSegmentationFactory)


class SegmentationStrategy(NamedTuple):
    name: str
    modality: str
    factory: Callable


class SegmentationStrategies(enum.Enum):

    DICOM_SEG = SegmentationStrategy(
        name="dicom_seg",
        modality="SEG",
        factory=DicomSEGSegmentationFactory
    )

    RT_STRUCT = SegmentationStrategy(
        name="rt_struct",
        modality="RTSTRUCT",
        factory=RTStructSegmentationFactory
    )

    def __init__(self, *args):
        """
        Used to make sure that the segmentation strategies enumeration contains no strategy with the same modality.
        """
        super().__init__()
        cls = self.__class__

        if any(self.value.modality == member.value.modality for member in cls):
            first_member_name = self.name
            first_member_modality_value = self.value.modality
            second_member_name = cls(self.value).name
            second_member_modality_value = cls(self.value).value.modality

            raise ValueError(
                f"Aliases not allowed in the SegmentationStrategies Enum Class. The value assigned to the "
                f"{first_member_name} attribute is {first_member_modality_value} and it is the same as the one "
                f"assigned to the {second_member_name} attribute, since it is also {second_member_modality_value}."
            )

    @classmethod
    def get_available_modalities(cls) -> List[str]:
        """
        Available segmentation DICOM modalities.

        Returns
        -------
        available_modalities : List[str]
            Available modalities.
        """
        return [member.value.modality for member in cls]



import setuptools

setuptools.setup(
    name="delia",
    version="1.1.1",
    author="Maxence Larose",
    author_email="maxence.larose.1@ulaval.ca",
    description="DICOM Extraction for Large-scale Image Analysis (DELIA).",
    long_description=open('README.md', "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/MaxenceLarose/delia",
    license="Apache License 2.0",
    keywords='dicom hdf5 medical image segmentation pre-processing python3 radiomics deep-learning dicom-seg rt-struct',
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "h5py",
        "monai==1.0.1",
        "numpy",
        "pandas",
        "pydicom",
        "pydicom-seg",
        "rt_utils",
        "scikit-image",
        "SimpleITK"
    ],
)



"""
    @file:              resample.py
    @Author:            Maxence Larose

    @Creation Date:     10/2022
    @Last modification: 11/2022

    @Description:       This file contains the Resampled transform.
"""

from typing import Callable, Dict, Hashable, Tuple

import SimpleITK as sitk
import numpy as np

from delia.transforms.physical_space.transform import PhysicalSpaceTransform, ImageData, KeysCollection, Mode


class Resampled(PhysicalSpaceTransform):
    """
    Resample an itk_image to new out_spacing.
    """

    def __init__(
            self,
            keys: KeysCollection,
            out_spacing: Tuple[float, float, float] = (1.0, 1.0, 1.0),
            interpolator: Callable = sitk.sitkBSpline,
    ):
        """
        Initialize output spacing.

        Parameters
        ----------
        keys : KeysCollection
            Keys of the corresponding items to be transformed. Image keys are assumed to be arbitrary series keys
            defined in 'series_descriptions'. For the label maps, the keys are organ names. Note that if
            'series_descriptions' is None, the image keys are assumed to be modality names.
        out_spacing : Tuple[int, int, int], default = (1.0, 1.0, 1.0)
            The desired spacing in the physical space. Default = (1.0 mm, 1.0 mm, 1.0 mm).
        interpolator : Callable
            The interpolator to be used for resampling the images. Default = sitk.sitkBSpline.
        """
        super().__init__(keys=keys)
        self._out_spacing = out_spacing
        self._interpolator = interpolator

    def __call__(self, data: Dict[str, ImageData]) -> Dict[Hashable, sitk.Image]:
        """
        Resample an itk_image to new out_spacing.

        Parameters
        ----------
        data : Dict[str, ImageData]
            A Python dictionary that contains ImageData.

        Returns
        -------
        transformed_image : Dict[Hashable, sitk.Image]
            A Python dictionary that contains transformed SimpleITK images.
        """
        d = dict(data)

        for key in self.key_iterator(d):
            original_itk_image = d[key].simple_itk_image

            original_spacing = original_itk_image.GetSpacing()
            original_size = original_itk_image.GetSize()

            out_size = [
                int(np.round(original_size[0] * (original_spacing[0] / self._out_spacing[0]))),
                int(np.round(original_size[1] * (original_spacing[1] / self._out_spacing[1]))),
                int(np.round(original_size[2] * (original_spacing[2] / self._out_spacing[2])))
            ]

            resample = sitk.ResampleImageFilter()
            resample.SetOutputSpacing(self._out_spacing)
            resample.SetSize(out_size)
            resample.SetOutputDirection(original_itk_image.GetDirection())
            resample.SetOutputOrigin(original_itk_image.GetOrigin())
            resample.SetTransform(sitk.Transform())
            resample.SetDefaultPixelValue(original_itk_image.GetPixelIDValue())

            if self._mode == Mode.NONE:
                raise AssertionError("Transform mode must be set before __call__.")
            elif self._mode == Mode.IMAGE:
                resample.SetInterpolator(self._interpolator)
            elif self._mode == Mode.SEGMENTATION:
                resample.SetInterpolator(sitk.sitkNearestNeighbor)
            else:
                raise ValueError("Unknown transform mode.")

            resampled_image = resample.Execute(original_itk_image)

            d[key] = resampled_image

        return d


ResampleD = ResampleDict = Resampled



"""
    @file:              patient_data_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains the PatientDataReader class which is used to read dicom files AND
                        segmentation files.
"""

import logging
from typing import Dict, List, Optional, Union

from monai.transforms import Compose
from monai.transforms import MapTransform as MonaiMapTransform

from delia.readers.image.dicom_reader import DicomReader
from delia.readers.patient_data.patient_data_query_context import PatientDataQueryContext
from delia.transforms.applications import apply_transforms
from delia.transforms.data.transform import DataTransform
from delia.transforms.physical_space.transform import PhysicalSpaceTransform
from delia.utils.data_model import PatientDataModel
from delia.utils.transforms_history import TransformsHistory

_logger = logging.getLogger(__name__)


class PatientDataReader(DicomReader):
    """
    A class used to read dicom files AND segmentation files.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            series_descriptions: Optional[Dict[str, List[str]]],
            erase_unused_dicom_files: bool = False
    ):
        """
        Used to check availability of given series' uid and series descriptions in the patient's dicom files.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's files.
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        super().__init__(path_to_patient_folder=path_to_patient_folder)

        self._images_dicom_headers = self.get_dicom_headers(remove_segmentations=True)
        self._series_descriptions = series_descriptions
        self._erase_unused_dicom_files = erase_unused_dicom_files

        self.failed_images = []
        if series_descriptions is not None:
            self.check_availability_of_given_series_description()

    @property
    def patient_id(self) -> str:
        """
        Patient ID.

        Returns
        -------
        patient_id : str
            Patient id.
        """
        return str(self._images_dicom_headers[0].PatientID)

    @property
    def paths_to_segmentations(self) -> List[str]:
        """
        Paths to segmentations.

        Returns
        -------
        paths_to_segmentations : List[str]
            List of paths to the segmentation files.
        """
        paths_to_segmentations = [
            series.paths_to_dicoms_from_series[0] for series in self._segmentations_series_data_dict.values()
        ]

        return paths_to_segmentations

    @property
    def series_descriptions(self) -> Dict[str, List[str]]:
        """
        Series descriptions setter.

        Returns
        -------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        """
        return self._series_descriptions

    @series_descriptions.setter
    def series_descriptions(self, series_descriptions: Dict[str, List[str]]):
        """
        Series descriptions setter.

        Parameters
        ----------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        """
        items = list(series_descriptions.items())
        for previous_items, current_items in zip(items, items[1:]):
            set_intersection = set(previous_items[1]) & set(current_items[1])

            if bool(set_intersection):
                raise AssertionError(
                    f"\nThe dictionary of series descriptions should not contain the same series names for different "
                    f"images/modalities. \nHowever, here we find the series names {previous_items[1]} for the "
                    f"{previous_items[0]} image and {current_items[1]} for the {current_items[0]} image. \nClearly, "
                    f"the images series values are overlapping because of the series named {set_intersection}."
                )

        self.series_descriptions = series_descriptions

    @property
    def available_series_descriptions(self) -> List[str]:
        """
        Available series descriptions.

        Returns
        -------
        available_series_descriptions : List[str]
            Available series descriptions in the patient dicom files.
        """
        available_series_descriptions = [
            self._get_series_description(dicom_header) for dicom_header in self._images_dicom_headers
        ]

        return available_series_descriptions

    def check_availability_of_given_series_description(self) -> None:
        """
        Check availability of given series description in the patient's dicom files.
        """
        _logger.debug("Checking availability of given series description...")
        for series_key, series_description_list in self.series_descriptions.items():
            if any(series in self.available_series_descriptions for series in series_description_list):
                pass
            else:
                self.record_failed_images(series_key)
        _logger.debug("Done.")

    def record_failed_images(self, series_key: str) -> None:
        """
        Record failed images.

        Parameters
        ----------
        series_key : str
            Series key.
        """
        _logger.error(
            f"Patient with ID {self.patient_id} has no series available that correlates with the image '{series_key}'. "
            f"The expected series descriptions for this image are {self.series_descriptions[series_key]} while the "
            f"patient record only contains the following series descriptions: {self.available_series_descriptions}."
        )

        self.failed_images.append(series_key)

    def get_patient_dataset(
            self,
            transforms: Optional[Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]] = None
    ) -> PatientDataModel:
        """
        Get the patient dataset.

        Parameters
        ----------
        transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
            A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on
            the SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that
            represents the image. DataTransform transforms the data using other a patient's other images or
            segmentations. The keys for images are assumed to be the arbitrary series key set in 'series_descriptions'.
            For segmentation, keys are organ names. Note that if 'series_descriptions' is None, the keys for images are
            assumed to be modalities.

        Returns
        -------
        patient_dataset : PatientDataModel
            A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
            segmentation data extracted from the segmentation files.
        """
        patient_data_context = PatientDataQueryContext(
            path_to_patient_folder=self._path_to_patient_folder,
            paths_to_segmentations=self.paths_to_segmentations,
            series_descriptions=self._series_descriptions,
            erase_unused_dicom_files=self._erase_unused_dicom_files
        )
        patient_dataset = patient_data_context.create_patient_data()

        if transforms:
            apply_transforms(patient_dataset=patient_dataset, transforms=transforms)

        _logger.debug(
            f"Chosen patient data query strategy: '{patient_data_context.patient_data_query_strategy.name}'."
        )
        _logger.info(f"{len(patient_dataset.data)} images added to the patient dataset, namely: ")

        for image_and_segmentation_data in patient_dataset.data:
            image = image_and_segmentation_data.image
            segmentations = image_and_segmentation_data.segmentations

            modality = image.dicom_header.Modality
            series_description = image.dicom_header.SeriesDescription
            image_segmentation_available = True if segmentations else False

            segmented_organs = set()
            if segmentations:
                for segmentation in segmentations:
                    for organ in list(segmentation.simple_itk_label_maps.keys()):
                        segmented_organs.add(organ)

            _logger.info(f"Series Description : {series_description}")
            _logger.info(f"  Modality : {modality}")
            _logger.info(f"  Image Segmentation available: {image_segmentation_available}")

            if image_segmentation_available:
                _logger.info(f"  Segmented organs : {segmented_organs}")

        patient_dataset.transforms_history = TransformsHistory(transforms)

        return patient_dataset



"""
    @file:              tools.py
    @Author:            Maxence Larose

    @Creation Date:     03/2023
    @Last modification: 03/2023

    @Description:       This file contains some tools to help applying array space transforms on images.
"""

from typing import Tuple, Union

import numpy as np
from monai.data import MetaTensor

from ..tools import convert_to_numpy


def compute_centroid(array: Union[MetaTensor, np.ndarray]) -> Tuple[int, int, int]:
    """
    Computes centroid.

    Parameters
    ----------
    array : Union[MetaTensor, np.ndarray]
        Image tensor or array.

    Returns
    -------
    centroid : Tuple[int, int, int]
        Centroid.
    """
    np_array = convert_to_numpy(array)
    assert np.array_equal(np_array, np_array.astype(bool)), (
        "The array to compute the centroid on must be a binary numpy array."
    )
    binary_array = (np_array == 1)
    centroid = np.argwhere(binary_array).sum(0)/binary_array.sum()

    return centroid.astype(int)



"""
    @file:              patient_data_query_strategy.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 03/2022

    @Description:       This file contains the class PatientDataQueryStrategies that enumerates the available categories
                        for the queries a user can make to obtain a patient's data.
"""

from enum import Enum
from typing import NamedTuple, Callable

from .factories.patient_data_factories import DefaultPatientDataFactory, SeriesDescriptionPatientDataFactory


class PatientDataQueryStrategy(NamedTuple):
    name: str
    factory: Callable


class PatientDataQueryStrategies(Enum):

    DEFAULT = PatientDataQueryStrategy(
        name="Default",
        factory=DefaultPatientDataFactory
    )

    SERIES_DESCRIPTION = PatientDataQueryStrategy(
        name="Series description",
        factory=SeriesDescriptionPatientDataFactory
    )

